/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>

#include "soribada.h"
#include "core.h"
#include "interface.h"
#include "callbacks.h"
#include "mp3.h"

/* from interface.c */
extern GdkPixmap *check;
extern GdkBitmap *checkmask;
extern GdkPixmap *uncheck;
extern GdkBitmap *uncheckmask;

/* search tab begin */
void on_search_entry_activated (GtkEditable *editable, gpointer data) {
	if (GTK_WIDGET_IS_SENSITIVE(app->search_button))
		on_search_button_clicked (NULL, data);
}

void on_search_button_clicked (GtkButton *button, gpointer data) {
	if (app->is_login) {
		if (app->search_in_result) {
			char *query = gtk_entry_get_text(GTK_ENTRY(data));
			char *iquery = gtk_entry_get_text(GTK_ENTRY(app->search_inverse_entry));

			app->stop_search = TRUE;
			search_in_result(query, iquery);
			show_status(_("Search result %d"), \
				    GTK_CLIST(app->search_clist)->rows);
		} else {
			char *query = gtk_entry_get_text(GTK_ENTRY(data));

			if (strlen(query) <= 0)
				return;

			gtk_clist_clear(GTK_CLIST(app->search_clist));
			show_status(_("Search for %s"), query);

			app->nreplies = 0;
			app->stop_search = FALSE;
			app->online_list_updated = FALSE;
			gtk_progress_set_value(GTK_PROGRESS(app->search_progressbar), 0);
			start_search(query);
		}
	} else {
		if (app->login_window) {
			gtk_window_activate_focus(GTK_WINDOW(app->login_window));
		} else {
			GtkWidget *window;

			window = create_login_window();
			gtk_widget_show(window);
		}
	}
}

void on_stop_button_clicked (GtkButton *button, gpointer data) {
	app->stop_search = TRUE;
}

void on_search_in_result_button_toggled(GtkToggleButton *togglebutton, gpointer data) {
	app->search_in_result = !app->search_in_result;
	if (app->search_in_result) {
		gtk_widget_show_all(GTK_WIDGET(data));
		app->redraw_search_clist = FALSE;
	} else if (app->redraw_search_clist) {
		gtk_widget_hide_all(GTK_WIDGET(data));
		redraw_search_clist();
	} else {
		gtk_widget_hide_all(GTK_WIDGET(data));
	}
}

/* clist menu */
void on_search_clist_menu_download_activate(GtkMenuItem *menuitem, gpointer data) {
	int row;
	gchar *filepath;
	MP3 *item;
	Downitem *ditem;
	GList *selected;

	selected = GTK_CLIST(app->search_clist)->selection;
	while (selected != NULL) {
		row = (gint)(selected->data);
		item = (MP3 *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->search_clist), row);
		filepath = g_strdup_printf("%s/%s", app->download_dir, \
					   basename_for_windows(item->path));
		gtk_clist_unselect_row(GTK_CLIST(app->search_clist), row, 0);
		if (check_filename_for_download(filepath)) {
			/* search Ʈ   */
			gtk_clist_remove(GTK_CLIST(app->search_clist), row);
			remove_from_replies_table(item);

			ditem = download_item_new(item);
			/* downloading clist ߰Ѵ */
			add_to_downloading_clist(ditem);
			/* downloads ̺ ߰Ѵ */
			add_to_downloads_table(ditem);

			/* download   */
			start_download(ditem);
		}
		selected = GTK_CLIST(app->search_clist)->selection;
	}
}

void on_search_clist_menu_download_as_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->search_clist)->selection;
	if (selected != NULL) {
		GtkWidget *filesel;
		int row;
		MP3 *item;

		row = (gint)selected->data;
		item = (MP3 *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->search_clist), row);
		filesel = create_download_as_dialog(basename_for_windows(item->path));
		gtk_widget_show(filesel);
	}
}

void on_search_clist_menu_download_queue_activate(GtkMenuItem *menuitem, gpointer data) {
	int row;
	gchar *filepath;
	MP3 *item;
	Downitem *ditem;
	GList *selected;

	selected = GTK_CLIST(app->search_clist)->selection;
	while (selected != NULL) {
		row = (gint)(selected->data);
		item = (MP3 *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->search_clist), row);
		filepath = g_strdup_printf("%s/%s", app->download_dir, \
					   basename_for_windows(item->path));
		gtk_clist_unselect_row(GTK_CLIST(app->search_clist), row, 0);
		if (check_filename_for_download(filepath)) {
			/* search Ʈ   */
			gtk_clist_remove(GTK_CLIST(app->search_clist), row);
			remove_from_replies_table(item);

			ditem = download_item_new(item);
			ditem->flag = DL_ST_QUEUE;
			/* downloading clist ߰Ѵ */
			add_to_downloading_clist(ditem);

			/* downloads ̺ ߰Ѵ */
			if (app->download_queue == NULL) {
				add_to_downloads_table(ditem);
				add_to_download_queue(ditem);
				start_download(ditem);
			} else {
				add_to_downloads_table(ditem);
				add_to_download_queue(ditem);
			}
		}
		selected = GTK_CLIST(app->search_clist)->selection;
	}
}

void on_search_clist_menu_delete_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	GList *selected;

	selected = GTK_CLIST(app->search_clist)->selection;
	gtk_clist_freeze(GTK_CLIST(app->search_clist));
	while (selected != NULL) {
		row = (gint)selected->data;
		gtk_clist_unselect_row(GTK_CLIST(app->search_clist), row, 0);
		gtk_clist_remove(GTK_CLIST(app->search_clist), row);
		selected = GTK_CLIST(app->search_clist)->selection;
	}
	gtk_clist_thaw(GTK_CLIST(app->search_clist));
	show_status(_("Search result %d"), GTK_CLIST(app->search_clist)->rows);
}

void on_search_clist_menu_message_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->search_clist)->selection;
	if (selected != NULL) {
		GtkWidget *dlg;
		int row;
		MP3 *item;
		Message *msg;
		Messageitem *mitem;

		row = (gint)selected->data;
		item = (MP3 *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->search_clist),row);

		/* ⼭ mitem  create_send_message_dialog()
		 * ѱ. ߿ message sending  ű⼭ ޸𸮸
		 * Ѵ */
		msg = message_new_from_mp3(item);
		mitem = (Messageitem *)message_item_new(msg);

		dlg = create_send_message_dialog(mitem);
		gtk_widget_show(dlg);
	}
}

void on_search_clist_select_row(GtkWidget *clist, gint row, gint colum, GdkEventButton *event, gpointer data) {
	g_return_if_fail(event != NULL);

	if (event->type == GDK_2BUTTON_PRESS) {
		if (event->button == 1)
			on_search_clist_menu_download_activate(NULL, NULL);
	}
}

void on_search_clist_click_column(GtkWidget *clist, gint column, gpointer data) {
	static GtkSortType type[8] = {
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
	};

	if (column == 3 || column == 4 || column == 7)
		gtk_clist_set_compare_func(GTK_CLIST(clist), compare_by_num);
	else
		gtk_clist_set_compare_func(GTK_CLIST(clist), NULL);
	gtk_clist_set_sort_column(GTK_CLIST(clist), column);
	gtk_clist_set_sort_type(GTK_CLIST(clist), type[column]);
	gtk_clist_sort(GTK_CLIST(clist));

	if (type[column] == GTK_SORT_ASCENDING)
		type[column] = GTK_SORT_DESCENDING;
	else
		type[column] = GTK_SORT_ASCENDING;
}

gboolean on_search_clist_button_press(GtkWidget *widget, GdkEvent *event, gpointer data) {
	if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *)event;
		if (bevent->button == 3) {
			if (GTK_CLIST(widget)->selection != NULL) 
				gtk_menu_popup (GTK_MENU(data), \
						NULL, NULL, NULL, \
						NULL, \
						bevent->button, bevent->time);

			return TRUE;
		}
	}

	return FALSE;
}
/* search tab end */

/* download tab begin */
void on_downloading_clist_menu_stop_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		gint row;
		Downitem *ditem;
		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);
		ditem->command = DL_CMD_STOP;
	}
}

void on_downloading_clist_menu_pause_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		gint row;
		Downitem *ditem;
		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);
		ditem->command = DL_CMD_PAUSE;
	}
}

void on_downloading_clist_menu_resume_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		gint row;
		Downitem *ditem;

		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);

		if (ditem->flag != DL_ST_RUNNING) {
			if (g_list_find(app->download_queue, ditem)) {
				/* change bg color */
				GdkColor *color = &(app->downloading_clist->style->base[GTK_STATE_NORMAL]);
				gtk_clist_set_background(GTK_CLIST(app->downloading_clist), row, color);
				remove_from_download_queue(ditem);
			}
			start_download(ditem);
		}
	}
}

void on_downloading_clist_menu_queueing_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		gint row;
		Downitem *ditem;

		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);

		if (ditem->flag != DL_ST_RUNNING && 
		    g_list_find(app->download_queue, ditem) == NULL) {
			GdkColor color = { 0, 0xe6e6, 0xe6e6, 0xfafa };

			add_to_download_queue(ditem);
			gtk_clist_set_background(GTK_CLIST(app->downloading_clist), row, &color);
			if (app->download_queue->next == NULL)
				start_download(ditem);
		}
	}
}

void free_data(gpointer data, gpointer user_data) {
	g_free(data);
}

void on_downloading_clist_menu_play_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	gchar *filename;
	Downitem *ditem;
	GList *selected;
	GList *playlist = NULL;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	while (selected != NULL) {

		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);

		filename = g_strconcat(ditem->path, DOWNLOAD_TAIL, NULL);
		playlist = g_list_append(playlist, filename);
		selected = selected->next;
	}
	if (playlist != NULL)  {
		play_mp3_file(playlist);
		g_list_foreach(playlist, free_data, NULL);
		g_list_free(playlist);
	}
}

void on_downloading_clist_menu_remove_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		gint row;
		Downitem *ditem;

		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);
		gtk_clist_remove(GTK_CLIST(app->downloading_clist), row);
		remove_from_download_queue(ditem);
		if (ditem->flag == DL_ST_RUNNING) 
			ditem->command = DL_CMD_REMOVE;
		else {
			remove_from_downloads_table(ditem);
			download_item_unref(ditem);
		}
	}
}

void on_downloading_clist_menu_delete_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		gint row;
		Downitem *ditem;

		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);
		gtk_clist_remove(GTK_CLIST(app->downloading_clist), row);
		remove_from_download_queue(ditem);
		if (ditem->flag == DL_ST_RUNNING) 
			ditem->command = DL_CMD_DELETE;
		else {
			gchar *filename = 
				g_strconcat(ditem->path, DOWNLOAD_TAIL, NULL);
			unlink(filename);
			g_free(filename);
			remove_from_downloads_table(ditem);
			download_item_unref(ditem);
		}
	}
}

void on_downloading_clist_menu_message_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloading_clist)->selection;
	if (selected != NULL) {
		GtkWidget *dlg;
		int row;
		Downitem *ditem;
		Message *msg;
		Messageitem *mitem;

		row = (gint)selected->data;
		ditem = (Downitem *) \
		        gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist),row);

		/* ⼭ mitem  create_send_message_dialog()
		 * ѱ. ߿ message sending  ű⼭ ޸𸮸
		 * Ѵ */
		msg = message_new_from_mp3(ditem->item);
		mitem = (Messageitem *)message_item_new(msg);

		dlg = create_send_message_dialog(mitem);
		gtk_widget_show(dlg);
	}
}

gboolean on_downloading_clist_button_press(GtkWidget *widget, GdkEvent *event, gpointer data) {
	if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *)event;
		if (bevent->button == 3) {
			GList *selected = GTK_CLIST(app->downloading_clist)->selection;
			if (selected != NULL) {
				gint row;
				Downitem *ditem;

				row = (gint)selected->data;
				ditem = (Downitem *) \
					gtk_clist_get_row_data(GTK_CLIST(app->downloading_clist), row);
				if (ditem->flag == DL_ST_RUNNING) {
					gtk_widget_set_sensitive(app->downloading_clist_resume_menu, FALSE);
					gtk_widget_set_sensitive(app->downloading_clist_queueing_menu, FALSE);
				} else {
					gtk_widget_set_sensitive(app->downloading_clist_resume_menu, TRUE);
					gtk_widget_set_sensitive(app->downloading_clist_queueing_menu, TRUE);
				}
				gtk_menu_popup (GTK_MENU (data), \
						NULL, NULL, NULL, \
						NULL, bevent->button, \
						bevent->time);
				return TRUE;
			}
		}
	}

	return FALSE;
}

void on_downloaded_clist_menu_play_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	Downitem *ditem;
	GList *selected;
	GList *playlist = NULL;

	selected = GTK_CLIST(app->downloaded_clist)->selection;
	while (selected != NULL) {
		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloaded_clist),
					       row);

		playlist = g_list_append(playlist, ditem->path);
		selected = selected->next;
	}

	if (playlist != NULL) {
		play_mp3_file(playlist);
		g_list_free(playlist);
	}
}

void on_downloaded_clist_menu_file_info_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloaded_clist)->selection;
	if (selected != NULL) {
		gint row;
		File_info *info;
		Downitem *ditem;
		GtkWidget *window;

		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloaded_clist), row);

		info = file_info_new();
		info->ditem = ditem;
		info->path = ditem->path;
		window = create_file_info_window(info);
		if (window != NULL)
			gtk_widget_show(window);
	}
}

void on_downloaded_clist_menu_remove_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	Downitem *ditem;
	GList *selected;

	selected = GTK_CLIST(app->downloaded_clist)->selection;
	while (selected != NULL) {
		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloaded_clist), row);
		gtk_clist_unselect_row(GTK_CLIST(app->downloaded_clist), row, 0);
		gtk_clist_remove(GTK_CLIST(app->downloaded_clist), row);
		ditem = remove_from_downloads_table(ditem);
		download_item_destroy(ditem);
		selected = GTK_CLIST(app->downloaded_clist)->selection;
	}
}

void on_downloaded_clist_menu_delete_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	Downitem *ditem;
	gchar *filename;
	GList *selected;

	selected = GTK_CLIST(app->downloaded_clist)->selection;
	while (selected != NULL) {
		row = (gint)selected->data;
		ditem = (Downitem *) \
			gtk_clist_get_row_data(GTK_CLIST(app->downloaded_clist), row);
		gtk_clist_unselect_row(GTK_CLIST(app->downloaded_clist), row, 0);
		gtk_clist_remove(GTK_CLIST(app->downloaded_clist), row);

		ditem = remove_from_downloads_table(ditem);
		if (ditem->flag != DL_ST_COMPLETED) {
			filename = g_strconcat(ditem->path, DOWNLOAD_TAIL, NULL);
			unlink(filename);
			g_free(filename);
		} else
			unlink(ditem->path);
		download_item_destroy(ditem);
		selected = GTK_CLIST(app->downloaded_clist)->selection;
	}
}

void on_downloaded_clist_menu_message_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->downloaded_clist)->selection;
	if (selected != NULL) {
		GtkWidget *dlg;
		int row;
		Downitem *ditem;
		Message *msg;
		Messageitem *mitem;

		row = (gint)selected->data;
		ditem = (Downitem *) \
		        gtk_clist_get_row_data(GTK_CLIST(app->downloaded_clist),row);

		/* ⼭ mitem  create_send_message_dialog()
		 * ѱ. ߿ message sending  ű⼭ ޸𸮸
		 * Ѵ */
		msg = message_new_from_mp3(ditem->item);
		mitem = (Messageitem *)message_item_new(msg);

		dlg = create_send_message_dialog(mitem);
		gtk_widget_show(dlg);
	}
}
/* download tab end */

/* upload tab start */
void on_upload_clist_menu_message_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->upload_clist)->selection;
	if (selected != NULL) {
		GtkWidget *dlg;
		int row;
		Upitem *uitem;
		Message *msg;
		Messageitem *mitem;

		row = (gint)selected->data;
		uitem = (Upitem *) \
		        gtk_clist_get_row_data(GTK_CLIST(app->upload_clist),row);

		/* ⼭ mitem  create_send_message_dialog()
		 * ѱ. ߿ message sending  ű⼭ ޸𸮸
		 * Ѵ */
		msg = message_new_from_mtp(uitem->mtp);
		mitem = (Messageitem *)message_item_new(msg);

		dlg = create_send_message_dialog(mitem);
		gtk_widget_show(dlg);
	}
}
/* upload tab end */

/* local file tab start */
void on_local_clist_menu_play_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	MP3db *item;
	GList *selected;
	GList *playlist = NULL;

	selected = GTK_CLIST(app->local_clist)->selection;
	while (selected != NULL) {
		row = (gint)selected->data;
		item = (MP3db *) \
			gtk_clist_get_row_data(GTK_CLIST(app->local_clist), row);

		playlist = g_list_append(playlist, item->realpath);
		selected = selected->next;
	}
	if (playlist != NULL) {
		play_mp3_file(playlist);
		g_list_free(playlist);
	}
}

void on_local_clist_menu_file_info_activate(GtkMenuItem *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->local_clist)->selection;
	if (selected != NULL) {
		gint row;
		File_info *info;
		MP3db *item;
		GtkWidget *window;

		row = (gint)selected->data;
		item = (MP3db *) \
			gtk_clist_get_row_data(GTK_CLIST(app->local_clist), row);

		info = file_info_new();
		info->mp3db = item;
		info->path = item->realpath;
		window = create_file_info_window(info);
		gtk_widget_show(window);
	}
}

void on_local_clist_menu_remove_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	MP3db *item;
	GList *selected;

	gtk_clist_freeze(GTK_CLIST(app->local_clist));
	selected = GTK_CLIST(app->local_clist)->selection;
	while (selected != NULL) {
		row = (gint)selected->data;
		item = (MP3db *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->local_clist), row);
		gtk_clist_unselect_row(GTK_CLIST(app->local_clist), row, 0);
		gtk_clist_remove(GTK_CLIST(app->local_clist), row);

		app->mp3_list = g_list_remove(app->mp3_list, item);
		mp3db_item_free(item);
		selected = GTK_CLIST(app->local_clist)->selection;
	}
	gtk_clist_thaw(GTK_CLIST(app->local_clist));
}

void on_local_clist_menu_delete_activate(GtkMenuItem *menuitem, gpointer data) {
	gint row;
	MP3db *item;
	GList *selected;

	gtk_clist_freeze(GTK_CLIST(app->local_clist));
	selected = GTK_CLIST(app->local_clist)->selection;
	while (selected != NULL) {
		row = (gint)selected->data;

		item = (MP3db *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->local_clist), row);
		gtk_clist_unselect_row(GTK_CLIST(app->local_clist), row, 0);
		gtk_clist_remove(GTK_CLIST(app->local_clist), row);

		app->mp3_list = g_list_remove(app->mp3_list, item);
		unlink(item->realpath);
		mp3db_item_free(item);
		selected = GTK_CLIST(app->local_clist)->selection;
	}
	gtk_clist_thaw(GTK_CLIST(app->local_clist));
}

void on_add_button_clicked (GtkButton *button, gpointer data) {
	GtkWidget *filesel;

	filesel = create_file_add_dialog(app->download_dir);
	gtk_widget_show(filesel);
}

void on_file_add_dlg_ok (GtkButton *button, gpointer data) {
	gchar *filename;
	MP3db *item;

	filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(data));
	if (is_mp3(filename)) {
		item = mp3db_list_search_by_realpath(app->mp3_list, filename);
		if (item == NULL) {
			char *text[6];
			GList *last;
			GList *newlist;
			gint row;
			MP3 *mp3;

			newlist = mp3db_list_append(app->mp3_list, filename);
			if (newlist == NULL)
				return;

			app->mp3_list = newlist;
			last = g_list_last(app->mp3_list);
			item = (MP3db *)last->data;
			mp3 = item->data;

			text[0] = g_basename(item->realpath);
			text[1] = mp3->title;
			text[2] = mp3->artist;
			text[3] = mp3->size;
			text[4] = mp3->bitrate;
			text[5] = mp3->time;
			row = gtk_clist_append(GTK_CLIST(app->local_clist), text);
			gtk_clist_set_row_data(GTK_CLIST(app->local_clist), row, item);
		}
	}
	
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_file_add_dlg_cancel (GtkButton *button, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_delete_button_clicked (GtkButton *button, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->local_clist)->selection;
	if (selected != NULL) {
		gint row;
		MP3db *item;
		row = (gint)selected->data;

		item = (MP3db *) \
		       gtk_clist_get_row_data(GTK_CLIST(app->local_clist), row);
		gtk_clist_remove(GTK_CLIST(app->local_clist), row);

		app->mp3_list = g_list_remove(app->mp3_list, item);
		mp3db_item_free(item);
	}
}

void on_scan_button_clicked (GtkButton *button, gpointer data) { 
	gchar *text[6];
	gint row;
	/*
	GdkCursor *cursor;
	GdkCursor *newcursor;
	*/
	GList *list;

	/*
	newcursor = gdk_cursor_new(GDK_WATCH);
	gdk_window_set_cursor(app->window->window, newcursor);
	*/

	gtk_clist_clear(GTK_CLIST(app->local_clist));
	scan_local_mp3_dir();
	list = app->mp3_list;
	while (list != NULL) {
		text[0] = g_basename(((MP3db *)(list->data))->realpath);
		text[1] = ((MP3db *)(list->data))->data->title;
		text[2] = ((MP3db *)(list->data))->data->artist;
		text[3] = ((MP3db *)(list->data))->data->size;
		text[4] = ((MP3db *)(list->data))->data->bitrate;
		text[5] = ((MP3db *)(list->data))->data->time;

		row = gtk_clist_append(GTK_CLIST(app->local_clist), text);
		gtk_clist_set_row_data(GTK_CLIST(app->local_clist), row, list->data);
		list = list->next;
	}
	/*
	gdk_cursor_destroy(newcursor);

	cursor = gdk_cursor_new(GDK_LEFT_PTR);
	gdk_window_set_cursor(app->window->window, cursor);
	gdk_window_set_cursor(app->window->window, NULL);
	gdk_cursor_destroy(cursor);
	*/
}

void on_delete_all_button_clicked (GtkButton *button, gpointer data) { 
	delete_mp3db_list();
	gtk_clist_clear(GTK_CLIST(app->local_clist));
}

void on_stop_playing_button_clicked (GtkButton *button, gpointer data) { 
	stop_mp3_playing();
}

void on_local_clist_click_column(GtkWidget *clist, gint column, gpointer data) {
	static GtkSortType type[6] = {
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING,
		GTK_SORT_ASCENDING
	};

	if (column == 3 || column == 4)
		gtk_clist_set_compare_func(GTK_CLIST(clist), compare_by_num);
	else
		gtk_clist_set_compare_func(GTK_CLIST(clist), NULL);
	gtk_clist_set_sort_column(GTK_CLIST(clist), column);
	gtk_clist_set_sort_type(GTK_CLIST(clist), type[column]);
	gtk_clist_sort(GTK_CLIST(clist));

	if (type[column] == GTK_SORT_ASCENDING)
		type[column] = GTK_SORT_DESCENDING;
	else
		type[column] = GTK_SORT_ASCENDING;
}

void on_clist_select_row (GtkCList *clist, gint row, gint column, GdkEvent *event, gpointer data) {

}

void on_clist_unselect_row (GtkCList *clist, gint row, gint column, GdkEvent *event, gpointer data) {

}

/* menu , download tab, upload tab, local file tab */
gboolean on_clist_button_press (GtkWidget *widget, GdkEvent *event, gpointer data) {
	if (event->type == GDK_BUTTON_PRESS) {
		GdkEventButton *bevent = (GdkEventButton *)event;
		if (bevent->button == 3) {
			if (GTK_CLIST(widget)->selection != NULL) 
				gtk_menu_popup (GTK_MENU (data), \
						NULL, NULL, NULL, \
						NULL, \
						bevent->button, bevent->time);
			return TRUE;
		}
	}

	return FALSE;
}

/* main window */
gboolean on_window_delete_event (GtkWidget *widget, GdkEvent *event, gpointer data) {
        gdk_window_get_size(app->window->window, &app->width, &app->height);
	gtk_widget_destroy(app->window);
	gtk_main_quit();

	quit_soribada();
	return FALSE;
}

gboolean on_window_destroy_event (GtkWidget *widget, GdkEvent *event, gpointer data) {
        gdk_window_get_size(app->window->window, &app->width, &app->height);
	gtk_widget_destroy(app->window);
	gtk_main_quit();

	quit_soribada();
	return FALSE;
}

/* menu callback */
void on_file_activate (GtkMenuItem *menuitem, gpointer data) {

}

void on_connect_to_server_activate (GtkMenuItem *menuitem, gpointer data) {
	if (app->login_window) {
		gtk_window_activate_focus(GTK_WINDOW(app->login_window));
	} else {
		GtkWidget *window;

		window = create_login_window();
		gtk_widget_show(window);
	}
}

void on_disconnect_activate (GtkMenuItem *menuitem, gpointer data) {
	wait_for_upload_thread();
	stop_server_process();
	reset_search();

	app->is_login = FALSE;
	gtk_widget_set_sensitive (app->menu_connect_to_server, TRUE);
	gtk_widget_set_sensitive (app->menu_disconnect, FALSE);
	show_status(_("Disonnected from soribada"));
}

void on_help_activate (GtkMenuItem *menuitem, gpointer data) {

}

void on_about_activate (GtkMenuItem *menuitem, gpointer data) {
	/*
	static gboolean pop = TRUE;
	GtkWidget *window;

	if (pop) {
		pop = FALSE;
		window = create_about_window();
		gtk_widget_show(window);
		gtk_object_set_data(GTK_OBJECT(window), "POP", &pop);
	}
	*/
	create_about_window();
}

void on_quit_activate (GtkMenuItem *menuitem, gpointer data) {
        gdk_window_get_size(app->window->window, &app->width, &app->height);
	gtk_main_quit();

	quit_soribada();
}

void on_soribada_activate (GtkMenuItem *menuitem, gpointer data) {

}

void on_menu_message_activate (GtkMenuItem *menuitem, gpointer data) {
	pop_message_window(NULL);
}

void on_preference_activate (GtkMenuItem *menuitem, gpointer data) {
	if (pref == NULL) {
		GtkWidget *window;

		pref = create_preferences();
		window = create_preference_window();
		gtk_widget_show(window);
	}
}

/* login window */
void on_login_username_entry_activated (GtkEditable *enditable, gpointer data) {
	gtk_widget_grab_focus(GTK_WIDGET(app->password_entry));
}

void on_login_password_entry_activated (GtkEditable *enditable, gpointer data) {
	gtk_signal_emit_by_name(GTK_OBJECT(data), "clicked", app->login_window);
}

void on_login_connect_button_clicked (GtkButton *button, gpointer data) {
	g_free(app->username);
	g_free(app->password);
	app->username = 
		g_strdup(gtk_entry_get_text(GTK_ENTRY(app->username_entry)));
	app->password = 
		g_strdup(gtk_entry_get_text(GTK_ENTRY(app->password_entry)));

	login(app->username, app->password);
	gtk_widget_destroy(GTK_WIDGET(data));
	app->login_window = NULL;
}

void on_login_cancel_button_clicked (GtkButton *button, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
	app->login_window = NULL;
}

/* alert window */
void on_alert_destroy (GtkWidget *widget, gpointer data) {
	gpointer *vbox = gtk_object_get_user_data(GTK_OBJECT(widget));
	*vbox = NULL;
	gtk_widget_destroy(GTK_WIDGET(widget));
}

void on_alert_ok_button_clicked (GtkButton *button, gpointer data) {
	gpointer *vbox = gtk_object_get_user_data(GTK_OBJECT(data));
	*vbox = NULL;
	gtk_widget_destroy(GTK_WIDGET(data));
}

/* preference window */
void on_preference_window_destroy (GtkObject *object, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_check_button_toggled(GtkToggleButton *togglebutton, gpointer data) {
	*((gboolean *)data) = !*((gboolean *)data);
}

void on_save_username_toggled (GtkToggleButton *togglebutton, gpointer data) { 
	gtk_widget_set_sensitive(GTK_WIDGET(data), !GTK_WIDGET_SENSITIVE(data));
	pref->save_username = !pref->save_username;
}

void on_save_password_toggled (GtkToggleButton *togglebutton, gpointer data) {
	gtk_widget_set_sensitive(GTK_WIDGET(data), !GTK_WIDGET_SENSITIVE(data));
	pref->save_password = !pref->save_password;
}

void on_default_server_toggled (GtkToggleButton *togglebutton, gpointer data) {
	gtk_widget_set_sensitive(GTK_WIDGET(data), !GTK_WIDGET_SENSITIVE(data));
	pref->use_default_server = !pref->use_default_server;
}

void on_use_proxy_server_toggled (GtkToggleButton *togglebutton, gpointer data) {
	gtk_widget_set_sensitive(GTK_WIDGET(data), !GTK_WIDGET_SENSITIVE(data));
	pref->use_proxy = !pref->use_proxy;
}

void on_download_dir_button_clicked (GtkButton *button, gpointer data) { 
	GtkWidget *filesel;

	filesel = create_select_download_dir_dialog(data);
	gtk_widget_show(filesel);
}

void on_mp3player_button_clicked (GtkButton *button, gpointer data) { 
	GtkWidget *filesel;

	filesel = create_select_mp3player_dialog(data);
	gtk_widget_show(filesel);
}

void on_use_bandwidth_button_toggled (GtkToggleButton *togglebutton, gpointer data) {
	gtk_widget_set_sensitive(GTK_WIDGET(data), !GTK_WIDGET_SENSITIVE(data));
	pref->use_bandwidth_limit = !pref->use_bandwidth_limit;
}

gboolean on_preferences_spinbutton_event (GtkWidget *widget, GdkEventExpose *event, gpointer data) {
	int *p = (int *)data;

	*p = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));

	return FALSE;
}

void on_local_mp3_dir_add_button_clicked (GtkButton *button, gpointer data) { 
	static GtkWidget *filesel = NULL;

	if (filesel == NULL) {
		filesel = create_select_local_mp3_dir_dialog(data);
		gtk_object_set_data(GTK_OBJECT(filesel), "SELF", &filesel);
		gtk_widget_show(filesel);
	}
}

void on_local_mp3_dir_delete_button_clicked (GtkButton *button, gpointer data) {
	GList *selected;
	GtkWidget *item;
	gint index;
	void *user_data;

	selected = GTK_LIST(data)->selection;
	if (selected == NULL)
		return;

	item = selected->data;
	user_data = gtk_object_get_user_data(GTK_OBJECT(item));
	pref->localmp3_dir = g_list_remove(pref->localmp3_dir, user_data);

	index = gtk_list_child_position(GTK_LIST(data), item);
	gtk_list_clear_items(GTK_LIST(data), index, index + 1);
}

void on_sound_device_menu_activated (GtkMenuItem *menuitem, gpointer data) { 
	pref->sound_device = (Sound_Device) gtk_object_get_user_data(GTK_OBJECT(menuitem));
	if (pref->sound_device == NO_SOUND)
		gtk_widget_set_sensitive(GTK_WIDGET(data), FALSE);
	else
		gtk_widget_set_sensitive(GTK_WIDGET(data), TRUE);
}

void on_sound_event_clist_selected (GtkCList *clist, gint row, gint col, GdkEventButton *event, gpointer data) {
	gtk_widget_set_sensitive(GTK_WIDGET(data), TRUE);
}

void on_sound_event_clist_unselected (GtkCList *clist, gint row, gint col, GdkEventButton *event, gpointer data) {
	gtk_widget_set_sensitive(GTK_WIDGET(data), FALSE);
}

gboolean on_sound_event_clist_button_press (GtkWidget *widget, GdkEventButton *event, gpointer data) {
	gint x = (gint)event->x;
	gint y = (gint)event->y;
	gint row, col;

	if (gtk_clist_get_selection_info(GTK_CLIST(data), x, y, &row, &col)) {
		gboolean *bool = (gboolean *) \
				gtk_clist_get_row_data(GTK_CLIST(data), row);
		if (col == 0) {
			*bool = !*bool;
			if (*bool)
				gtk_clist_set_pixmap(GTK_CLIST(data), row, 0, \
						check, checkmask);
			else
				gtk_clist_set_pixmap(GTK_CLIST(data), row, 0, \
						uncheck, uncheckmask);
			return TRUE;
		}
	}

	return FALSE;
}

void on_sound_test_button_clicked (GtkButton *button, gpointer data) { 
	gchar *file;
	gint row;

	if (GTK_CLIST(data)->selection == NULL)
		return;

	row = GPOINTER_TO_INT(GTK_CLIST(data)->selection->data);
	if (gtk_clist_get_text(GTK_CLIST(data), row, 2, &file) == 1)
		play_wave_file(file, pref->sound_device);
}

void on_sound_modify_button_clicked (GtkButton *button, gpointer data) { 
	GtkWidget *dlg;

	dlg = create_select_sound_dialog(data);
	if (dlg != NULL) {
		gtk_widget_show(dlg);
		gtk_grab_add(dlg);
	}
}

void on_select_sound_dlg_ok (GtkButton *button, gpointer data) { 
	gchar *newpath;
	gint row;
	gboolean *bool;
	GtkCList *clist;

	clist = GTK_CLIST(gtk_object_get_user_data(GTK_OBJECT(data)));
	if (clist->selection == NULL)
		return;

	row = GPOINTER_TO_INT(clist->selection->data);
	bool = (gboolean *)gtk_clist_get_row_data(clist, row);
	newpath = gtk_file_selection_get_filename(GTK_FILE_SELECTION(data));
	gtk_clist_set_text(clist, row, 2, newpath);
	
	if (bool == &pref->use_sound_on_search) {
		g_free(pref->sound_on_search);
		pref->sound_on_search = g_strdup(newpath);
	} else if (bool == &pref->use_sound_on_down) {
		g_free(pref->sound_on_down);
		pref->sound_on_down = g_strdup(newpath);
	} else if (bool == &pref->use_sound_on_message) {
		g_free(pref->sound_on_message);
		pref->sound_on_message = g_strdup(newpath);
	}

	gtk_grab_remove(GTK_WIDGET(data));
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_select_sound_dlg_cancel (GtkButton *button, gpointer data) { 
	gtk_grab_remove(GTK_WIDGET(data));
	gtk_widget_destroy(GTK_WIDGET(data));
}

gboolean on_preferences_entry_event (GtkWidget *widget, GdkEventExpose *event, gpointer data) {
	char **p = (char **)data;

	g_free(*p);
	*p = g_strdup_printf(gtk_entry_get_text(GTK_ENTRY(widget)));

	return FALSE;
}

void on_preference_ok_button_clicked (GtkButton *button, gpointer data) { 
	gtk_widget_destroy(GTK_WIDGET(data));

        gdk_pixmap_unref(check);
        gdk_bitmap_unref(checkmask);
        gdk_pixmap_unref(uncheck);
        gdk_bitmap_unref(uncheckmask);

	/* replies table   
	 * ̰ app->maxreplies ʿϱ  pref ϱ 
	 * ۾Ѵ */
	reallocate_replies_table(pref->maxreplies);

	copy_preferences_to_app(pref);
	destroy_preferences(pref);
	pref = NULL;

	/* proxy setting  */
	setup_proxy_setting(app->proxy_server);

	/* search clist  */
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 1, \
                                         app->search_clist_title);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 2, \
                                         app->search_clist_artist);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 3, \
                                         app->search_clist_comment);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 4, \
                                         app->search_clist_size);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 5, \
                                         app->search_clist_bitrate);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 6, \
                                         app->search_clist_time);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 7, \
                                         app->search_clist_user);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 8, \
                                         app->search_clist_speed);
        gtk_clist_set_column_visibility (GTK_CLIST (app->search_clist), 0, \
                                         app->search_clist_filename);
	/* downloading clist  */
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloading_clist), 1, \
                                         app->downloading_clist_user);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloading_clist), 2, \
                                         app->downloading_clist_size);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloading_clist), 3, \
                                         app->downloading_clist_received);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloading_clist), 4, \
                                         app->downloading_clist_progress);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloading_clist), 0, \
                                         app->downloading_clist_filename);

	/* downloaded clist  */
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloaded_clist), 1, \
                                         app->downloaded_clist_user);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloaded_clist), 2, \
                                         app->downloaded_clist_size);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloaded_clist), 3, \
                                         app->downloaded_clist_received);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloaded_clist), 4, \
                                         app->downloaded_clist_status);
        gtk_clist_set_column_visibility (GTK_CLIST (app->downloaded_clist), 0, \
                                         app->downloaded_clist_filename);
}

void on_preference_cancel_button_clicked (GtkButton *button, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
	destroy_preferences(pref);
	pref = NULL;

        gdk_pixmap_unref(check);
        gdk_bitmap_unref(checkmask);
        gdk_pixmap_unref(uncheck);
        gdk_bitmap_unref(uncheckmask);
}

void on_download_as_ok_clicked (GtkButton *button, gpointer data) { 
	GList *selected;

	selected = GTK_CLIST(app->search_clist)->selection;
	if (selected != NULL) {
		int row;
		gchar *filepath;
		MP3 *item;
		Downitem *ditem;

		filepath = gtk_file_selection_get_filename(GTK_FILE_SELECTION(data));
		if (check_filename_for_download(filepath)) {
			/* search Ʈ   */
			row = (gint)selected->data;
			item = (MP3 *) \
			       gtk_clist_get_row_data(GTK_CLIST(app->search_clist), row);

			gtk_clist_remove(GTK_CLIST(app->search_clist), row);
			remove_from_replies_table(item);

			ditem = download_item_new(item);
			ditem->path = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(data)));

			/* downloading Ʈ ߰Ѵ */
			add_to_downloading_clist(ditem);
			/* downloads ̺ ߰Ѵ */
			add_to_downloads_table(ditem);

			/* download   */
			start_download(ditem);
		} else
			return;
	}
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_download_as_cancel_clicked (GtkButton *button, gpointer data) { 
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_select_download_dir_dlg_ok (GtkButton *button, gpointer data) {
	GtkWidget *entry;

	entry = GTK_WIDGET(gtk_object_get_user_data(GTK_OBJECT(data)));
	gtk_entry_set_text(GTK_ENTRY(entry), \
			   gtk_file_selection_get_filename(GTK_FILE_SELECTION(data)));
	gtk_widget_destroy(GTK_WIDGET(data));
	gtk_widget_grab_focus(entry);
}

void on_select_download_dir_dlg_cancel (GtkButton *button, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_select_mp3player_dlg_ok (GtkButton *button, gpointer data) {
	GtkWidget *entry;

	entry = GTK_WIDGET(gtk_object_get_user_data(GTK_OBJECT(data)));
	gtk_entry_set_text(GTK_ENTRY(entry), \
			   gtk_file_selection_get_filename(GTK_FILE_SELECTION(data)));
	gtk_widget_destroy(GTK_WIDGET(data));
	gtk_widget_grab_focus(entry);
}

void on_select_mp3player_dlg_cancel (GtkButton *button, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
}

static gint compare(gconstpointer a, gconstpointer b) {
	return strncmp((const char *)a, (const char *)b, 1024);
}

void on_select_local_mp3_dir_dlg_ok (GtkButton *button, gpointer data) {
	GtkWidget *list;
	GtkWidget *listitem;
	GtkWidget **window;
	gchar *path;

	path = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(data)));
	if (g_list_find_custom(pref->localmp3_dir, path, &compare) == NULL) {
		list = gtk_object_get_user_data(GTK_OBJECT(data));
		listitem = gtk_list_item_new_with_label(path);
		gtk_widget_show(listitem);
		gtk_object_set_user_data(GTK_OBJECT(listitem), path);
		gtk_container_add(GTK_CONTAINER(list), listitem);

		pref->localmp3_dir = g_list_append(pref->localmp3_dir, path);
	}

	window = gtk_object_get_data(GTK_OBJECT(data), "SELF");
	gtk_widget_destroy(GTK_WIDGET(data));
	*window = NULL;
}

void on_select_local_mp3_dir_dlg_cancel (GtkButton *button, gpointer data) {
	GtkWidget **window = gtk_object_get_data(GTK_OBJECT(data), "SELF");
	gtk_widget_destroy(GTK_WIDGET(data));
	*window = NULL;
}

void on_about_button_clicked (GtkButton *button, gpointer data) { 
	gtk_widget_destroy(GTK_WIDGET(data));
}

/* file info window */
void on_file_info_entry_changed(GtkEditable *editable, gpointer data) {
	if (!GTK_WIDGET_IS_SENSITIVE(data))
		gtk_widget_set_sensitive(GTK_WIDGET(data), TRUE);
}

void on_file_info_genre_selected(GtkList *list, GtkWidget *widget, \
				 gpointer data) {
	((File_info *)data)->genre_no = \
			(gint)gtk_object_get_data(GTK_OBJECT(widget), "genre");
}

void on_use_id3_tag_button_toggled(GtkToggleButton *togglebutton, \
				   gpointer data) {
	File_info *info = (File_info *)data;

	if (gtk_toggle_button_get_active(togglebutton)) {
		info->use_id3_tag = TRUE;
		gtk_widget_set_sensitive(info->title, TRUE);
		gtk_widget_set_sensitive(info->artist, TRUE);
		gtk_widget_set_sensitive(info->album, TRUE);
		gtk_widget_set_sensitive(info->comment, TRUE);
		gtk_widget_set_sensitive(info->year, TRUE);
		gtk_widget_set_sensitive(info->track, TRUE);
		gtk_widget_set_sensitive(info->genre, TRUE);
	} else {
		info->use_id3_tag = FALSE;
		gtk_widget_set_sensitive(info->title, FALSE);
		gtk_widget_set_sensitive(info->artist, FALSE);
		gtk_widget_set_sensitive(info->album, FALSE);
		gtk_widget_set_sensitive(info->comment, FALSE);
		gtk_widget_set_sensitive(info->year, FALSE);
		gtk_widget_set_sensitive(info->track, FALSE);
		gtk_widget_set_sensitive(info->genre, FALSE);
	}
	gtk_widget_set_sensitive(GTK_WIDGET(info->save_button), TRUE);
}

void on_file_info_save_button_clicked (GtkButton *button, gpointer data) { 
	int ret;
	File_info *info = (File_info *)data;

	if (info->use_id3_tag) {
		/* ID3 ±׸  */
		ret = mp3_write_id3tag(info->path,
			gtk_entry_get_text(GTK_ENTRY(info->title)),
			gtk_entry_get_text(GTK_ENTRY(info->artist)),
			gtk_entry_get_text(GTK_ENTRY(info->album)),
			gtk_entry_get_text(GTK_ENTRY(info->comment)),
			gtk_entry_get_text(GTK_ENTRY(info->year)),
			strtol(gtk_entry_get_text(GTK_ENTRY(info->track)), \
				NULL, 10),
			info->genre_no);
		if (ret < 0) {
			alert(_("Can't update ID3v1 tag: %s"), strerror(errno));
			return;
		}
		if (info->mp3db != NULL) {
			/* local clist ̴   */
			gint row;
			MP3 *mp3;

			row = gtk_clist_find_row_from_data( \
					GTK_CLIST(app->local_clist),\
					info->mp3db);
			if (row < 0) {
				gtk_widget_set_sensitive(GTK_WIDGET(button), \
							 FALSE);
				return;
			}


			gtk_clist_set_text(GTK_CLIST(app->local_clist), row, \
				1, gtk_entry_get_text(GTK_ENTRY(info->title)));
			gtk_clist_set_text(GTK_CLIST(app->local_clist), row, \
				2, gtk_entry_get_text(GTK_ENTRY(info->artist)));

			/* mp3db   */
			mp3 = info->mp3db->data;
			g_free(mp3->title);
			mp3->title = g_strdup(gtk_entry_get_text( \
						GTK_ENTRY(info->title)));
			g_free(mp3->artist);
			mp3->artist = g_strdup(gtk_entry_get_text( \
						GTK_ENTRY(info->artist)));
			g_free(mp3->cmt);
			mp3->cmt = g_strdup(gtk_entry_get_text( \
						GTK_ENTRY(info->comment)));
		}
	} else {
		/* ID3 ±׸  */
		ret = mp3_remove_id3tag(info->path);
		if (ret < 0) {
			alert(_("Can't delete ID3v1 tag: %s"), strerror(errno));
			return;
		}
		if (info->mp3db != NULL) {
			/* local clist ̴   */
			gint row;
			MP3 *mp3;

			row = gtk_clist_find_row_from_data( \
					GTK_CLIST(app->local_clist),\
					info->mp3db);
			if (row < 0) {
				gtk_widget_set_sensitive(GTK_WIDGET(button), \
							 FALSE);
				return;
			}

			gtk_clist_set_text(GTK_CLIST(app->local_clist), row, \
					   1, "");
			gtk_clist_set_text(GTK_CLIST(app->local_clist), row, \
					   2, "");

			/* mp3db   */
			mp3 = info->mp3db->data;
			g_free(mp3->title);
			mp3->title = g_strdup("");
			g_free(mp3->artist);
			mp3->artist = g_strdup("");
			g_free(mp3->cmt);
			mp3->cmt = g_strdup("");
		}
	}

	gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);
}

void on_file_info_rename_button_clicked (GtkWidget *widget, gpointer data) { 
	int row;
	char *oldname, *newname;
	File_info *info = (File_info *)data;

	if (info == NULL)
		return;

	/*  â  â ƴϱ   â   Ҹ޾â
	 * Ͽ ִ     ع  ü
	 *  ؼ seg fault Եȴ. ׷   ׻  ü
	 * ִ ˻縦 ؾ Ѵ. */
	if (info->ditem != NULL) {
		/* ٿε    */
		row = gtk_clist_find_row_from_data(\
					GTK_CLIST(app->downloaded_clist),
					info->ditem);
		if (row < 0) {
			gtk_widget_set_sensitive(GTK_WIDGET(info->rename_button),
						 FALSE);
			return;
		}

		oldname = info->ditem->path;
		newname = g_strdup_printf("%s/%s", g_dirname(oldname),
				gtk_entry_get_text(GTK_ENTRY(info->filename)));
		if (is_file_exist(newname)) {
			alert(_("%s\nis already exist\nUse another file name"), 
			      newname); 
			gtk_widget_set_sensitive(GTK_WIDGET(info->rename_button),
				      		 FALSE);
			return;
		}
		if (rename(oldname, newname) == 0) {
			g_free(info->ditem->path);
			info->ditem->path = newname;

			gtk_clist_set_text(GTK_CLIST(app->downloaded_clist), \
					   row, 0, g_basename(newname));
		} else {
			alert("rename:%s", strerror(errno)); 
		}
	} else if (info->mp3db != NULL) {
		/*  ũ   */
		row = gtk_clist_find_row_from_data(GTK_CLIST(app->local_clist),
						   info->mp3db);
		if (row < 0) {
			gtk_widget_set_sensitive(GTK_WIDGET(info->rename_button),
						 FALSE);
			return;
		}

		oldname = info->mp3db->realpath;
		newname = g_strdup_printf("%s/%s", g_dirname(oldname),
				gtk_entry_get_text(GTK_ENTRY(info->filename)));
		if (is_file_exist(newname)) {
			alert(_("%s\nis already exist\nUse another file name"), 
			      newname); 
			gtk_widget_set_sensitive(GTK_WIDGET(info->rename_button),
						 FALSE);
			return;
		}
		if (rename(oldname, newname) == 0) {
			g_free(info->mp3db->realpath);
			info->mp3db->realpath = newname;

			gtk_clist_set_text(GTK_CLIST(app->local_clist), \
					   row, 0, g_basename(newname));
		} else {
			alert("rename:%s", strerror(errno)); 
		}
	}

	gtk_widget_set_sensitive(GTK_WIDGET(info->rename_button), FALSE);
}

void on_file_info_close_button_clicked (GtkButton *button, gpointer data) { 
	File_info *info = gtk_object_get_data(GTK_OBJECT(data), "file_info");

	gtk_widget_destroy(GTK_WIDGET(data));
	file_info_free(info);
}


/* message window */
void on_message_window_close_button_clicked (GtkButton *button, gpointer data) {
	app->message_window = NULL;
	gtk_widget_destroy(GTK_WIDGET(data));
}

void on_message_clist_menu_reply_activate(GtkButton *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->message_clist)->selection;
	if (selected != NULL) {
		GtkWidget *dlg;
		int row;
		Message *msg;
		Messageitem *mitem;
		Messageitem *msgitem;

		row = (gint)selected->data;
		msgitem = (Messageitem *) gtk_clist_get_row_data( \
					GTK_CLIST(app->message_clist), row);

		/* ⼭ mitem ؼ create_send_message_dialog()
		 * ѱ. ߿ message sending  ű⼭ ޸𸮸
		 * Ѵ */
		msg = message_new();
		msg->username = g_strdup(msgitem->msg->username);
		msg->addr = g_strdup(msgitem->msg->addr);
		msg->port = msgitem->msg->port;
		mitem = (Messageitem *)message_item_new(msg);

		dlg = create_send_message_dialog(mitem);
		gtk_widget_show(dlg);
	}
}

void on_message_clist_menu_delete_activate(GtkButton *menuitem, gpointer data) {
	GList *selected;

	selected = GTK_CLIST(app->message_clist)->selection;
	if (selected != NULL) {
		int row;
		guint len;
		Messageitem *mitem;

		row = (gint)selected->data;
		mitem = (Messageitem *) gtk_clist_get_row_data( \
					GTK_CLIST(app->message_clist), row);

		app->messages = g_list_remove(app->messages, mitem);
		message_item_free(mitem);
		gtk_clist_remove(GTK_CLIST(app->message_clist), row);

		gtk_text_freeze(GTK_TEXT(app->message_text));
		len = gtk_text_get_length(GTK_TEXT(app->message_text));
		gtk_text_backward_delete(GTK_TEXT(app->message_text), len);
		gtk_text_thaw(GTK_TEXT(app->message_text));
	}
}

void on_message_clist_select_row(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer data) {
	Messageitem *mitem;

	mitem = (Messageitem *) \
		gtk_clist_get_row_data(GTK_CLIST(app->message_clist), row);
	set_message_text(GTK_WIDGET(data), mitem);
}

void on_message_send_button_clicked (GtkButton *button, gpointer data) {
	GtkEditable *text;
	Messageitem *mitem;

	text = (GtkEditable *)gtk_object_get_data(GTK_OBJECT(data), "Text");
	mitem = (Messageitem *)gtk_object_get_data(GTK_OBJECT(data), \
						   "Messageitem");

	mitem->msg->text = g_strdup(gtk_editable_get_chars(text, 0, -1));
	gtk_widget_destroy(GTK_WIDGET(data));

	start_message_send(mitem);
}

void on_message_close_button_clicked (GtkButton *button, gpointer data) {
	Messageitem *mitem;

	mitem = (Messageitem *)gtk_object_get_data(GTK_OBJECT(data), \
						   "Messageitem");
	message_item_free(mitem);
	gtk_widget_destroy(GTK_WIDGET(data));
}
