/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <glib.h>

#include "soribada.h"
#include "core.h"

int str_to_int(gchar *value) {
	return strtol(value, NULL, 10);
}

gboolean str_to_boolean(gchar *value) {
	if (g_strcasecmp(value, "TRUE") == 0)
		return TRUE;
	else
		return FALSE;
}

gchar * boolean_to_str(gboolean value) {
	if (value)
		return "TRUE";
	else
		return "FALSE";
}

int parse_conf_file(char *name) {
	char *key;
	char *val;
	char buf[256];
	GList *newlist = NULL;
	GList *list = NULL;
	FILE *file;

	file = fopen(name, "r");
	if (file == NULL) {
		perror(_("Can't read config file, load default value"));
		return -1;
	}

	while (fgets(buf, sizeof (buf), file)) {
		if (strcmp(buf, "[General]\n") == 0) {
			while (fgets(buf, sizeof (buf), file)) {
				if (buf[0] == '[' || buf[0] == '\n')
					break;
				key = strtok(buf, "=");
				val = strtok(NULL, "\n");
				if (strcmp(key, "username") == 0) {
					if (val != NULL)
						app->username = g_strdup(val);
				} else if (strcmp(key, "password") == 0) {
					if (val != NULL)
						app->password = g_strdup(val);
				} else if (strcmp(key, "server") == 0) {
					if (val != NULL) {
						g_free(app->soribada_server);
						app->soribada_server = g_strdup(val);
					}
				} else if (strcmp(key, "proxy_server") == 0) {
					if (val != NULL) {
						g_free(app->proxy_server);
						app->proxy_server = g_strdup(val);
					}
				} else if (strcmp(key, "download_dir") == 0) {
					if (val != NULL) {
						g_free(app->download_dir);
						app->download_dir = g_strdup(val);
					}
				} else if (strcmp(key, "localmp3_dir") == 0) {
					if (val != NULL)
						list = g_list_append(list, g_strdup(val));
				} else if (strcmp(key, "port") == 0) {
					if (val != NULL)
						app->port = str_to_int(val);
				} else if (strcmp(key, "check_online") == 0) {
					if (val != NULL) 
						app->check_online = str_to_boolean(val);
				} else if (strcmp(key, "use_proxy") == 0) {
					if (val != NULL) 
						app->use_proxy = str_to_boolean(val);
				} else if (strcmp(key, "mp3player") == 0) {
					if (val != NULL) {
						g_free(app->mp3player);
						app->mp3player = g_strdup(val);
					}
				} else if (strcmp(key, "auto_login") == 0) {
					app->auto_login = str_to_boolean(val);
				} else if (strcmp(key, "save_username") == 0) {
					app->save_username = str_to_boolean(val);
				} else if (strcmp(key, "save_password") == 0) {
					app->save_password = str_to_boolean(val);
				} else if (strcmp(key, "use_default_server") == 0) {
					app->use_default_server = str_to_boolean(val);
				} else if (strcmp(key, "user_limit") == 0) {
					if (val != NULL)
						app->user_limit = str_to_int(val);
				} else if (strcmp(key, "use_bandwidth_limit") == 0) {
					app->use_bandwidth_limit = str_to_boolean(val);
				} else if (strcmp(key, "bandwidth_limit") == 0) {
					if (val != NULL)
						app->bandwidth_limit = str_to_int(val);
				} else {
					fprintf(stderr, _("Unknown keyword %s, ignore\n"), key);
				}
			}
		} else if (strcmp(buf, "[GUI]\n") == 0) {
			while (fgets(buf, sizeof (buf), file)) {
				if (buf[0] == '[' || buf[0] == '\n')
					break;
				key = strtok(buf, "=");
				val = strtok(NULL, "\n");
				if (strcmp(key, "show_query_list") == 0) {
					app->show_query_list = str_to_boolean(val);
				} else if (strcmp(key, "no_of_max_replies") == 0) {
					app->maxreplies = str_to_int(val);
				} else if (strcmp(key, "width") == 0) {
					app->width = str_to_int(val);
				} else if (strcmp(key, "height") == 0) {
					app->height = str_to_int(val);
				} else if (strcmp(key, "sound_device") == 0) {
					if (strcmp(val, "no_sound") == 0)
						app->sound_device = NO_SOUND;
					else if (strcmp(val, "beep_sound") == 0)
						app->sound_device = BEEP_SOUND;
					else if (strcmp(val, "esd_sound") == 0)
						app->sound_device = ESD_SOUND;
					else
						app->sound_device = NO_SOUND;
				} else if (strcmp(key, "sound_on_search") == 0) {
					if (val != NULL) {
						g_free(app->sound_on_search);
						app->sound_on_search = g_strdup(val);
					}
				} else if (strcmp(key, "sound_on_down") == 0) {
					if (val != NULL) {
						g_free(app->sound_on_down);
						app->sound_on_down = g_strdup(val);
					}
				} else if (strcmp(key, "sound_on_message") == 0) {
					if (val != NULL) {
						g_free(app->sound_on_message);
						app->sound_on_message = g_strdup(val);
					}
				} else if (strcmp(key, "use_sound_on_search") == 0) {
					app->use_sound_on_search = str_to_boolean(val);
				} else if (strcmp(key, "use_sound_on_down") == 0) {
					app->use_sound_on_down = str_to_boolean(val);
				} else if (strcmp(key, "use_sound_on_message") == 0) {
					app->use_sound_on_message = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_filename") == 0) {
					app->search_clist_filename = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_title") == 0) {
					app->search_clist_title = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_artist") == 0) {
					app->search_clist_artist = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_comment") == 0) {
					app->search_clist_comment = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_size") == 0) {
					app->search_clist_size = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_bitrate") == 0) {
					app->search_clist_bitrate = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_time") == 0) {
					app->search_clist_time = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_user") == 0) {
					app->search_clist_user = str_to_boolean(val);
				} else if (strcmp(key, "search_clist_speed") == 0) {
					app->search_clist_speed = str_to_boolean(val);
				} else if (strcmp(key, "downloading_clist_filename") == 0) {
					app->downloading_clist_filename = str_to_boolean(val);
				} else if (strcmp(key, "downloading_clist_user") == 0) {
					app->downloading_clist_user = str_to_boolean(val);
				} else if (strcmp(key, "downloading_clist_size") == 0) {
					app->downloading_clist_size = str_to_boolean(val);
				} else if (strcmp(key, "downloading_clist_received") == 0) {
					app->downloading_clist_received = str_to_boolean(val);
				} else if (strcmp(key, "downloading_clist_progress") == 0) {
					app->downloading_clist_progress = str_to_boolean(val);
				} else if (strcmp(key, "downloading_clist_speed") == 0) {
					app->downloading_clist_speed = str_to_boolean(val);
				} else if (strcmp(key, "downloaded_clist_filename") == 0) {
					app->downloaded_clist_filename = str_to_boolean(val);
				} else if (strcmp(key, "downloaded_clist_user") == 0) {
					app->downloaded_clist_user = str_to_boolean(val);
				} else if (strcmp(key, "downloaded_clist_size") == 0) {
					app->downloaded_clist_size = str_to_boolean(val);
				} else if (strcmp(key, "downloaded_clist_received") == 0) {
					app->downloaded_clist_received = str_to_boolean(val);
				} else if (strcmp(key, "downloaded_clist_status") == 0) {
					app->downloaded_clist_status = str_to_boolean(val);
				} else if (strcmp(key, "scan_sub_dirs") == 0) {
					app->scan_sub_dirs = str_to_boolean(val);
				} else {
					fprintf(stderr, _("Unknown keyword %s, ignore\n"), key);
				}
			}
		}
	}

	newlist = list;
	list = app->localmp3_dir;
	while (list != NULL) {
		g_free(list->data);
		list = list->next;
	}
	g_list_free(app->localmp3_dir);
	app->localmp3_dir = newlist;

	fclose(file);

	return 0;
}

void load_default_value(void) {
	app->username = NULL;
	app->password = NULL;

	app->soribada_server = g_strdup(DEFAULT_SERVER);
	app->proxy_server = NULL;
	app->download_dir = g_strdup(getenv("HOME"));
	app->localmp3_dir = NULL;
	app->localmp3_dir = g_list_append(app->localmp3_dir, \
					  g_strdup(getenv("HOME")));
	app->port = DEFAULT_PORT;
	app->check_online = TRUE;
	app->use_proxy = FALSE;

	app->mp3_list_file = g_strdup_printf("%s/%s", getenv("HOME"), MP3_LIST);

	app->user_limit = 3;
	app->maxreplies = DEFAULT_MAX_REPLIES;
	app->bandwidth_limit = 100;

	app->mp3player = g_strdup(DEFAULT_MP3_PLAYER);
	app->width = 600;
	app->height = 400;
	app->auto_login = FALSE;
	app->save_username = FALSE;
	app->save_password = FALSE;
	app->use_default_server = TRUE;
	app->use_bandwidth_limit = FALSE;
	app->show_query_list = FALSE;

	app->sound_device = NO_SOUND;
	app->sound_on_search = g_strdup_printf("%s/%s", \
					SOUNDDIR, DEFAULT_SOUND_SEARCH);
	app->sound_on_down = g_strdup_printf("%s/%s", \
					SOUNDDIR, DEFAULT_SOUND_DOWN);
	app->sound_on_message = g_strdup_printf("%s/%s", \
					SOUNDDIR, DEFAULT_SOUND_MESSAGE);
	app->use_sound_on_search = TRUE;
	app->use_sound_on_down = TRUE;
	app->use_sound_on_message = TRUE;

	app->search_clist_filename = TRUE;
	app->search_clist_title = TRUE;
	app->search_clist_artist = TRUE;
	app->search_clist_comment = TRUE;
	app->search_clist_size = TRUE;
	app->search_clist_bitrate = TRUE;
	app->search_clist_time = TRUE;
	app->search_clist_user = TRUE;
	app->search_clist_speed = TRUE;

	app->downloading_clist_filename = TRUE;
	app->downloading_clist_user = TRUE;
	app->downloading_clist_size = TRUE;
	app->downloading_clist_received = TRUE;
	app->downloading_clist_progress = TRUE;
	app->downloading_clist_speed = TRUE;

	app->downloaded_clist_filename = TRUE;
	app->downloaded_clist_user = TRUE;
	app->downloaded_clist_size = TRUE;
	app->downloaded_clist_received = TRUE;
	app->downloaded_clist_status = TRUE;

	app->scan_sub_dirs = FALSE;
}

/* conf  о app  Ѵ */
void load_conf_file(void) {
	char *conf_file;

	conf_file = g_strdup_printf("%s/%s", getenv("HOME"), CONF_FILE);
	if (access(conf_file, R_OK) == 0) {
		if (parse_conf_file(conf_file) < 0) {
			load_default_value();
		}
	} 
	g_free(conf_file);
}

void write_str_option(FILE *fp, char *name, char *value) {
	if (value != NULL)
		fprintf(fp, "%s=%s\n", name, value);
}

void write_int_option(FILE *fp, char *name, int value) {
	fprintf(fp, "%s=%d\n", name, value);
}

void write_boolean_option(FILE *fp, char *name, gboolean value) {
	if (value)
		fprintf(fp, "%s=TRUE\n", name);
	else
		fprintf(fp, "%s=FALSE\n", name);
}


void write_conf_file(void) {
	GList *list;
	char *conf_file;
	FILE *file;

	conf_file = g_strdup_printf("%s/%s", getenv("HOME"), CONF_FILE);

	file = fopen(conf_file, "w");
	if (file == NULL) {
		perror(_("Can't write config file"));
	} else {
		/* change permission */
		if (fchmod(fileno(file), S_IRUSR | S_IWUSR) < 0)
			perror(_("Can't change config file permission"));

		/* general section */
		fprintf(file, "[General]\n");
		if (app->username && app->save_username)
			write_str_option(file, "username", app->username);
		if (app->password && app->save_password)
			write_str_option(file, "password", app->password);

		write_str_option(file, "server", app->soribada_server);
		write_str_option(file, "download_dir", app->download_dir);
		list = app->localmp3_dir;
		while (list != NULL) {
			write_str_option(file, "localmp3_dir", (char *)list->data);
			list = list->next;
		}
		write_int_option(file, "port", app->port);
		write_boolean_option(file, "check_online", app->check_online);

		/* proxy setting */
		write_boolean_option(file, "use_proxy", app->use_proxy);
		write_str_option(file, "proxy_server", app->proxy_server);

		write_str_option(file, "mp3player", app->mp3player);
		write_boolean_option(file, "auto_login", app->auto_login);
		write_boolean_option(file, "save_username", app->save_username);
		write_boolean_option(file, "save_password", app->save_password);
		write_boolean_option(file, "use_default_server", app->use_default_server);
		write_int_option(file, "user_limit", app->user_limit);
		write_boolean_option(file, "use_bandwidth_limit", app->use_bandwidth_limit);
		write_int_option(file, "bandwidth_limit", app->bandwidth_limit);
		fprintf(file, "\n");

		/* GUI section */
		fprintf(file, "[GUI]\n");
		write_int_option(file, "no_of_max_replies", app->maxreplies);

		write_int_option(file, "width", app->width);
		write_int_option(file, "height", app->height);
		write_boolean_option(file, "show_query_list", app->show_query_list);

		switch (app->sound_device) {
		case NO_SOUND:
			write_str_option(file, "sound_device", "no_sound");
			break;
		case BEEP_SOUND:
			write_str_option(file, "sound_device", "beep_sound");
			break;
		case ESD_SOUND:
			write_str_option(file, "sound_device", "esd_sound");
			break;
		default:
			write_str_option(file, "sound_device", "no_sound");
			break;
		}

		write_str_option(file, "sound_on_search", app->sound_on_search);
		write_boolean_option(file, "use_sound_on_search", \
					app->use_sound_on_search);
		write_str_option(file, "sound_on_down", app->sound_on_down);
		write_boolean_option(file, "use_sound_on_down", \
					app->use_sound_on_down);
		write_str_option(file, "sound_on_message", app->sound_on_message);
		write_boolean_option(file, "use_sound_on_message", \
					app->use_sound_on_message);

		write_boolean_option(file, "search_clist_filename", \
			app->search_clist_filename);
		write_boolean_option(file, "search_clist_title", \
			app->search_clist_title);
		write_boolean_option(file, "search_clist_artist", \
			app->search_clist_artist);
		write_boolean_option(file, "search_clist_comment", \
			app->search_clist_comment);
		write_boolean_option(file, "search_clist_size", \
			app->search_clist_size);
		write_boolean_option(file, "search_clist_bitrate", \
			app->search_clist_bitrate);
		write_boolean_option(file, "search_clist_time", \
			app->search_clist_time);
		write_boolean_option(file, "search_clist_user", \
			app->search_clist_user);
		write_boolean_option(file, "search_clist_speed", \
			app->search_clist_speed);
		write_boolean_option(file, "downloading_clist_filename", \
			app->downloading_clist_filename);
		write_boolean_option(file, "downloading_clist_user", \
			app->downloading_clist_user);
		write_boolean_option(file, "downloading_clist_size", \
			app->downloading_clist_size);
		write_boolean_option(file, "downloading_clist_received", \
			app->downloading_clist_received);
		write_boolean_option(file, "downloading_clist_progress", \
			app->downloading_clist_progress);
		write_boolean_option(file, "downloading_clist_speed", \
			app->downloading_clist_speed);
		write_boolean_option(file, "downloaded_clist_filename", \
			app->downloaded_clist_filename);
		write_boolean_option(file, "downloaded_clist_user", \
			app->downloaded_clist_user);
		write_boolean_option(file, "downloaded_clist_size", \
			app->downloaded_clist_size);
		write_boolean_option(file, "downloaded_clist_received", \
			app->downloaded_clist_received);
		write_boolean_option(file, "downloaded_clist_status", \
			app->downloaded_clist_status);

		write_boolean_option(file, "scan_sub_dirs", app->scan_sub_dirs);

		fclose(file);
	}
	g_free(conf_file);
}

