/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>

#include "soribada.h"
#include "protocol.h"
#include "interface.h"
#include "core.h"

//static pthread_mutex_t downmutex = PTHREAD_MUTEX_INITIALIZER;
//static pthread_mutex_t upmutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t nusermutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t quit_cond = PTHREAD_COND_INITIALIZER;

static unsigned char msgbuf[4096];

void show_status(const gchar *fmt, ...) {
	static gint msg_id = -1;
	gchar *message;
	gchar *text;
	va_list arg;

	va_start(arg, fmt);
	text = g_strdup_vprintf(fmt, arg);
	va_end(arg);

	message = g_strconcat("  ", text, NULL);
	if (msg_id > 0)
		gtk_statusbar_pop(GTK_STATUSBAR (app->statusbar), \
				  app->context_id);
	msg_id = gtk_statusbar_push(GTK_STATUSBAR (app->statusbar), \
				    app->context_id, message);
	g_free(message);
	g_free(text);
}

/* download item */
Downitem *download_item_new(MP3 *item) {
	Downitem *ditem;

	ditem = (Downitem *)xmalloc(sizeof(Downitem));
	ditem->command = DL_CMD_NONE;
	ditem->flag = DL_ST_NONE;
	ditem->ref_count = 1;
	ditem->received = 0;
	ditem->speed = 0;
	ditem->path = g_strdup_printf("%s/%s", app->download_dir, \
				  basename_for_windows(item->path));
	ditem->status = NULL;
	ditem->item = item;

	return ditem;
}

void download_item_destroy(Downitem *ditem) {
	g_free(ditem->path);
	g_free(ditem->status);
	mp3_free(ditem->item);
	g_free(ditem);
}

void download_item_ref(Downitem *ditem) {
	if (ditem != NULL)
		ditem->ref_count++;
}

void download_item_unref(Downitem *ditem) {
	if (ditem != NULL) {
		ditem->ref_count--;
		if (ditem->ref_count <= 0)
			download_item_destroy(ditem);
	}
}

/* upload item */
Upitem *upload_item_new(void) {
	Upitem *uitem;

	uitem = (Upitem *)xmalloc(sizeof(Upitem));
	uitem->path = NULL;
	uitem->mtp = NULL;
	uitem->item = NULL;

	return uitem;
}

void upload_item_destroy(Upitem *uitem) {
	g_free(uitem->path);
	g_free(uitem);
}

Downitem *add_to_downloads_table(Downitem *ditem) {
	app->downloads = g_list_append(app->downloads, (gpointer)ditem);

	return ditem;
}

Downitem *add_to_download_queue(Downitem *ditem) {
	app->download_queue = g_list_append(app->download_queue, (gpointer)ditem);

	return ditem;
}

Downitem *find_in_downloads_table_by_path(gchar *path) {
	GList *list = app->downloads;

	while (list != NULL) {
		if (strcmp(path, ((Downitem *)list->data)->path) == 0)
			break;
		list = list->next;
	}
	if (list == NULL)
		return NULL;
	else 
		return list->data;
}

Downitem *remove_from_downloads_table(Downitem *ditem) {
	app->downloads = g_list_remove(app->downloads, ditem);

	return ditem;
}

Downitem *remove_from_download_queue(Downitem *ditem) {
	app->download_queue = g_list_remove(app->download_queue, ditem);

	return ditem;
}

void remove_from_replies_table(MP3 *item) {
	int i;

	for (i = 0; i < app->nreplies; i++) {
		if (app->replies[i] == item) {
			/* Լ downloads̺ item ѱؼ
			 * ϴ Լ̹Ƿ 
			 * ⼭ ޸𸮸   ʴ´ */
			app->replies[i] = NULL;
			break;
		}
	}
}

int add_to_downloading_clist(Downitem *ditem) {
	int row;
	char *text[6];
	GdkColor color = { 0, 0xe6e6, 0xe6e6, 0xfafa };

	text[0] = g_basename(ditem->path);
	text[1] = ditem->item->user;
	text[2] = ditem->item->size;
	text[3] = NULL;
	text[4] = NULL;
	if (ditem->flag == DL_ST_QUEUE)
		text[5] = _("Queued");
	else
		text[5] = NULL;

	row = gtk_clist_append(GTK_CLIST(app->downloading_clist), text);
	gtk_clist_set_row_data(GTK_CLIST(app->downloading_clist), row, ditem);
	if (ditem->flag == DL_ST_QUEUE)
		gtk_clist_set_background(GTK_CLIST(app->downloading_clist), row, &color);

	return row;
}

gboolean start_download_wrapper(gpointer data) {
	start_download((Downitem *)data);
	return FALSE;
}

void update_download_on_running(Downitem* ditem) {
	gint row;
	char buf[16];
	int filelen;
	float progress;

	row = gtk_clist_find_row_from_data(GTK_CLIST(app->downloading_clist),
					   ditem);
	snprintf(buf, sizeof(buf), "%d", ditem->received);
	gtk_clist_set_text(GTK_CLIST(app->downloading_clist), row, 3, buf);

	snprintf(buf, sizeof(buf), "%0.2f", ditem->speed);
	gtk_clist_set_text(GTK_CLIST(app->downloading_clist), row, 4, buf);

	filelen = strtol(ditem->item->size, NULL, 10);
	progress = (float)(ditem->received) / (float)filelen * 100.0;
	snprintf(buf, sizeof(buf), "%0.2f", progress);
	gtk_clist_set_text(GTK_CLIST(app->downloading_clist), row, 5, buf);
}

void update_download_on_complete(Downitem* ditem) {
	gint row;
	Downitem *qditem;
	char buf[16];
	char *text[5];

	/* download Ϸ */
	/* downloading list  */
	row = gtk_clist_find_row_from_data(GTK_CLIST(app->downloading_clist),
					   ditem);
	gtk_clist_remove(GTK_CLIST(app->downloading_clist), row);

	/* downloaded list ߰ */
	text[0] = g_basename(ditem->path);
	text[1] = ditem->item->user;
	text[2] = ditem->item->size;
	text[3] = NULL;
	text[4] = ditem->status;
	row = gtk_clist_append(GTK_CLIST(app->downloaded_clist), text);
	gtk_clist_set_row_data(GTK_CLIST(app->downloaded_clist), row, ditem);

	g_snprintf(buf, sizeof(buf), "%d", ditem->received);
	gtk_clist_set_text(GTK_CLIST(app->downloaded_clist), row, 3, buf);

	/* download_queue  ִ  Ȱȭ */
	if (app->download_queue != NULL && \
	    g_list_find(app->download_queue, ditem) != NULL) {
		/* ٿε尡 ϷǾ queue Ѵ 
		 * ׸ queue ִ° Ȯؾ   Ȱȭ
		 * Ų */
		remove_from_download_queue(ditem);
		if (app->download_queue != NULL) {
			qditem = (Downitem *)app->download_queue->data;
			start_download(qditem);
		}
	}
}

void update_download_on_stop(Downitem* ditem) {
	gint row;
	Downitem *qditem;
	char buf[16];
	char *text[5];

	/* download  */
	/* downloading list  */
	row = gtk_clist_find_row_from_data(GTK_CLIST(app->downloading_clist),
					   ditem);
	gtk_clist_remove(GTK_CLIST(app->downloading_clist), row);

	/* downloaded list ߰ */
	text[0] = g_basename(ditem->path);
	text[1] = ditem->item->user;
	text[2] = ditem->item->size;
	text[3] = NULL;
	text[4] = ditem->status;
	row = gtk_clist_append(GTK_CLIST(app->downloaded_clist), text);
	gtk_clist_set_row_data(GTK_CLIST(app->downloaded_clist), row, ditem);

	g_snprintf(buf, sizeof(buf), "%d", ditem->received);
	gtk_clist_set_text(GTK_CLIST(app->downloaded_clist), row, 3, buf);
	if (ditem->status == NULL)
		gtk_clist_set_text(GTK_CLIST(app->downloaded_clist),
				   row, 4, _("Stoped"));
	else
		gtk_clist_set_text(GTK_CLIST(app->downloaded_clist),
				   row, 4, ditem->status);

	/* download_queue  ִ  Ȱȭ */
	if (app->download_queue != NULL && \
	    g_list_find(app->download_queue, ditem) != NULL) {
		/* ٿε尡 ϷǾ queue Ѵ 
		 * ׸ queue ִ° Ȯؾ   Ȱȭ
		 * Ų */
		remove_from_download_queue(ditem);
		if (app->download_queue != NULL) {
			qditem = (Downitem *)app->download_queue->data;
			start_download(qditem);
		}
	}
}

void update_download_on_default(Downitem* ditem) {
	gint row;
	char buf[16];
	Downitem *qditem;

	/*    */
	/* Ͽ   */
	row = gtk_clist_find_row_from_data(GTK_CLIST(app->downloading_clist),
					   ditem);
	g_snprintf(buf, sizeof(buf), "%d", ditem->received);
	gtk_clist_set_text(GTK_CLIST(app->downloading_clist), row, 3, buf);
	g_snprintf(buf, sizeof(buf), "%0.2f", ditem->speed);
	gtk_clist_set_text(GTK_CLIST(app->downloading_clist), row, 4, buf);
	if (ditem->status == NULL)
		gtk_clist_set_text(GTK_CLIST(app->downloading_clist),
				   row, 5, _("Paused"));
	else
		gtk_clist_set_text(GTK_CLIST(app->downloading_clist),
				   row, 5, ditem->status);

	/* download_queue  ִ  Ȱȭ */
	if (app->download_queue != NULL && \
	    g_list_find(app->download_queue, ditem) != NULL) {
		/* ٿε尡 ϷǾ queue Ѵ 
		 * ׸ queue ִ° Ȯؾ   Ȱȭ
		 * Ų */
		int delay = 0;
		remove_from_download_queue(ditem);
		/* Queue ϳ ۿ   쿡 
		 * 5  ð  д */
		if (app->download_queue == NULL)
			delay = 0;
		/* Ѱ   ٽ  */
		add_to_download_queue(ditem);
		if (app->download_queue != NULL) {
			qditem = (Downitem *)app->download_queue->data;
			gtk_timeout_add(delay, &start_download_wrapper, qditem);
		}
	}
}

gboolean update_download(gpointer data) {
	Downitem *ditem = (Downitem *)data;

	switch (ditem->flag) {
	case DL_ST_FREED:
		remove_from_downloads_table(ditem);
		remove_from_download_queue(ditem);
		download_item_unref(ditem);
		return FALSE;

	case DL_ST_RUNNING:
		update_download_on_running(ditem);
		return TRUE;

	case DL_ST_COMPLETED:
		update_download_on_complete(ditem);
		return FALSE;

	case DL_ST_STOPED:
		update_download_on_stop(ditem);
		return FALSE;

	default:
		update_download_on_default(ditem);
		return FALSE;
	}
}

void *download_thread(void *data) {
	int fd;
	char buf[4096];
	MTP *mtp;
	Downitem *ditem = (Downitem *)data;

	download_item_ref(ditem);
	fd = connect_by_address_nb(ditem->item->addr, ditem->item->port, 30);
	if (fd < 0) {
		g_free(ditem->status);
		if (errno == ETIMEDOUT)
			ditem->status = strdup(_("Timed out"));
		else
			ditem->status = strdup(_("Network error"));
		download_item_unref(ditem);
		ditem->flag = DL_ST_STOPED;
		return NULL;
	}

	mtp = mtp_new(fd);
	mtp->filename = g_strdup(ditem->item->path);
	mtp->position = ditem->received;

	if (mtp_send_request(app->pl, mtp, buf, sizeof(buf)) < 0) {
		g_free(ditem->status);
		ditem->status = strdup(_("Request error"));
		close(fd);
		mtp_free(mtp);
		download_item_unref(ditem);
		ditem->flag = DL_ST_STOPED;
		return NULL;
	}

	if (mtp_recv_return_code(app->pl, mtp, buf, sizeof(buf)) < 0) {
		g_free(ditem->status);
		ditem->status = strdup(_("Request error"));
		close(fd);
		mtp_free(mtp);
		download_item_unref(ditem);
		ditem->flag = DL_ST_STOPED;
		return NULL;
	}

	if (mtp->code == 0) {
		/* now we start download mp3 data */
		int i, n, ret;
		fd_set rset;
		float recvd;
		char *filename;
		struct timeval s, c;
		struct timeval timeout;
		FILE *fp;

		filename = g_strconcat(ditem->path, DOWNLOAD_TAIL, NULL);
		if (mtp->position == 0)
			fp = fopen(filename, "w");
		else
			fp = fopen(filename, "r+");

		if (fp == NULL) {
			g_free(ditem->status);
			ditem->status = strdup(_("Local file open error"));
			perror(ditem->path);
			close(fd);
			mtp_free(mtp);
			download_item_unref(ditem);
			ditem->flag = DL_ST_STOPED;
			return NULL;
		}
		
		fseek(fp, mtp->position, SEEK_SET);
		mtp->received = mtp->position;
		ditem->received = mtp->position;

		i = 1;
		recvd = 0;
		gettimeofday(&s, NULL);
		FD_ZERO(&rset);
		while (mtp->received < mtp->size) {
			FD_SET(fd, &rset);
			timeout.tv_sec = 60;
			timeout.tv_usec = 0;
			ret = select(fd + 1, &rset, NULL, NULL, &timeout);
			if (ret < 0) {
				if (errno == EINTR)
					continue;
				perror("download:select");
				g_free(ditem->status);
				ditem->status = strdup(_("Network error"));
				close(fd);
				fclose(fp);
				mtp_free(mtp);
				download_item_unref(ditem);
				ditem->flag = DL_ST_PAUSED;
				return NULL;
			}
			if (ret == 0) {
				g_free(ditem->status);
				ditem->status = strdup(_("Timed out"));
				close(fd);
				fclose(fp);
				mtp_free(mtp);
				download_item_unref(ditem);
				ditem->flag = DL_ST_PAUSED;
				return NULL;
			}

			if (FD_ISSET(fd, &rset)) {
				n = read(fd, buf, sizeof(buf));

				if (n == 0) {
					g_free(ditem->status);
					ditem->status = strdup(_("Closed"));
					close(fd);
					fclose(fp);
					mtp_free(mtp);
					download_item_unref(ditem);
					ditem->flag = DL_ST_PAUSED;
					return NULL;
				}
				if (n < 0) {
					if (errno == EINTR)
						continue;
					perror("read");
					g_free(ditem->status);
					ditem->status = strdup(_("Network error"));
					close(fd);
					fclose(fp);
					mtp_free(mtp);
					download_item_unref(ditem);
					ditem->flag = DL_ST_PAUSED;
					return NULL;
				}

				fwrite(buf, n, 1, fp);
				mtp->received += n;
				ditem->received += n;
				recvd += n;
			}

			if (i % 3 == 0) {
				if (mtp->received == mtp->size)
					break;
				gettimeofday(&c, NULL);
				/* kbytes  */
				mtp->speed = recvd / \
					     ((float)(c.tv_sec - s.tv_sec) * 1000.0 + \
					      (float)(c.tv_usec - s.tv_usec) / 1000.0);
				ditem->speed = mtp->speed;
				mtp_send_progress(app->pl, mtp, buf, sizeof(buf));
			}
			if (ditem->command == DL_CMD_STOP ||
			    ditem->command == DL_CMD_PAUSE ||
			    ditem->command == DL_CMD_REMOVE ||
			    ditem->command == DL_CMD_DELETE)
			       	break;
			i++;
		}

		if (mtp->received == mtp->size)
			mtp_send_done(app->pl, mtp, buf, sizeof(buf));
		fclose(fp);
		close(fd);
		mtp_free(mtp);

		if (mtp->received == mtp->size) {
			rename(filename, ditem->path);
			g_free(ditem->status);
			ditem->status = strdup(_("Done"));
			download_item_unref(ditem);
			ditem->flag = DL_ST_COMPLETED;
			if (app->use_sound_on_down)
				sound_event(SOUND_ON_DOWN);
		} else if (ditem->command == DL_CMD_REMOVE) {
			download_item_unref(ditem);
			ditem->flag = DL_ST_FREED;
		} else if (ditem->command == DL_CMD_DELETE) {
			download_item_unref(ditem);
			unlink(filename);
			ditem->flag = DL_ST_FREED;
		} else if (ditem->command == DL_CMD_PAUSE) {
			download_item_unref(ditem);
			ditem->flag = DL_ST_PAUSED;
		} else if (ditem->command == DL_CMD_STOP) {
			download_item_unref(ditem);
			ditem->flag = DL_ST_STOPED;
		} else {
			download_item_unref(ditem);
			ditem->flag = DL_ST_STOPED;
		}
	} else if (mtp->code == MTP_USER_LIMIT) {
	       	close(fd);
		mtp_free(mtp);
		g_free(ditem->status);
		ditem->status = strdup(_("User limit"));
		download_item_unref(ditem);
		ditem->flag = DL_ST_PAUSED;
	} else if (mtp->code == MTP_DOWNLOAD_LIMIT) {
		close(fd);
		mtp_free(mtp);
		g_free(ditem->status);
		ditem->status = strdup(_("User limit"));
		download_item_unref(ditem);
		ditem->flag = DL_ST_PAUSED;
	} else if (mtp->code == MTP_FILE_NOT_FOUND) {
		close(fd);
		mtp_free(mtp);
		g_free(ditem->status);
		ditem->status = strdup(_("File not found"));
		download_item_unref(ditem);
		ditem->flag = DL_ST_PAUSED;
	} else if (mtp->code == MTP_INVALID_COMMAND) {
		close(fd);
		mtp_free(mtp);
		g_free(ditem->status);
		ditem->status = strdup(_("Invalid request"));
		download_item_unref(ditem);
		ditem->flag = DL_ST_PAUSED;
	} else {
		close(fd);
		mtp_free(mtp);
		g_free(ditem->status);
		ditem->status = g_strdup_printf("%s(%d)", mtp->msg, mtp->code);
		download_item_unref(ditem);
		ditem->flag = DL_ST_PAUSED;
	}
	return NULL;
}

void start_download(Downitem *ditem) {
	pthread_t tid;

	if (ditem->flag == DL_ST_RUNNING)
		return;

	if (ditem->path == NULL)
		ditem->path = g_strdup_printf("%s/%s", app->download_dir, \
				  basename_for_windows(ditem->item->path));

	ditem->command = DL_CMD_RUN;
	ditem->flag = DL_ST_RUNNING;
	gtk_timeout_add(500, &update_download, ditem);
	pthread_create(&tid, NULL, &download_thread, ditem);
	pthread_detach(tid);
}

/* search  Լ */
gboolean stop_search(gpointer data) {
	app->stop_search = TRUE;

	gtk_widget_set_sensitive(app->search_button, TRUE);
	gtk_progress_set_value(GTK_PROGRESS(app->search_progressbar), 0);
	show_status(_("Search result %d from %d users (Completed)"), \
				    app->nreplies, app->nonlines);
	if (app->use_sound_on_search)
		sound_event(SOUND_ON_SEARCH);
	return FALSE;
}

gboolean send_query_by_time(gpointer data) {
	if (app->online_list_updated) {
		static int id = 0;
		int fd;
		struct sockaddr_in addr;
		Packet *packet = (Packet *)data;

		if (app->listen_channel == NULL) {
			query_packet_free(packet);
			return FALSE;
		}

		if (id == 0) {
			/* search  */
			gtk_progress_configure(GTK_PROGRESS(app->search_progressbar),
						0.0, 0.0, app->nonlines);

			show_status(_("Search result %d from %d users"), \
				    app->nreplies, app->nonlines);
		}

		fd = g_io_channel_unix_get_fd(app->listen_channel);

		update_query_packet_id(packet, app->query_id);

		addr.sin_family = AF_INET;
		addr.sin_port = htons(app->onlines[id]->port);
		inet_pton(AF_INET, app->onlines[id]->addr, &addr.sin_addr);

		/* üũ 
		 *    Ŷ Ű澵 ʿ䰡  */
		sendto(fd, packet->data, packet->len, 0, \
				(struct sockaddr *)&addr, sizeof(addr));
		gettimeofday(&app->time[app->query_id], NULL);

		id++;
		app->query_id++;

		if (id < app->nonlines && \
		    app->nreplies < app->maxreplies && \
		    !app->stop_search) {
			gtk_progress_set_value(GTK_PROGRESS(app->search_progressbar), (float)id);
			gtk_progress_draw_text(GTK_PROGRESS(app->search_progressbar));
			return TRUE;
		} else {
			/* query packet ׸  */
			query_packet_free(packet);
			app->online_list_updated = FALSE;

			if (id >= app->nonlines)
				/* 2 ʰ  ٸ ˻  */
				gtk_timeout_add(2000, &stop_search, NULL);
			else
				/* ٷ ˻  */
				stop_search(NULL);
			id = 0;

			return FALSE; 	/* watch func  */
		}
	} else if (app->stop_search) {
		/* online Ʈ ޾ƿµ  */
		query_packet_free((Packet *)data);

		return FALSE;
	} else 
		return TRUE;
}

void broadcast_queries(const gchar *query_string) {
	Packet *packet;

	packet = query_packet_new(app->pl, query_string);
	/* g_io Լ  ʰ  gtk_timeout  ߴ.
	g_io_add_watch(app->listen_channel, G_IO_OUT, \
		       &send_query, (gpointer)packet);
		       */
	app->send_query_handler = gtk_timeout_add(5, &send_query_by_time, \
						  (gpointer)packet);
}

void *search_thread(void *data) {
	app->nonlines = get_online_list(app->pl, app->onlines);

	if (app->nonlines > 0)
		app->online_list_updated = TRUE;
	else
		app->stop_search = TRUE;
	return NULL;
}

void start_search(const gchar *query_string) {
	pthread_t tid;

	gtk_widget_set_sensitive(app->search_button, FALSE);
	pthread_create(&tid, NULL, &search_thread, (void*)query_string);
	pthread_detach(tid);

	show_status(_("Send queries for %s"), query_string);
	/* query  */
	broadcast_queries(query_string);
}

void reset_search(void) {
	if (!app->stop_search) {
		app->stop_search = TRUE;
		app->nreplies = 0;
		gtk_timeout_remove(app->send_query_handler);
		gtk_clist_clear(GTK_CLIST(app->search_clist));
		gtk_progress_set_value(GTK_PROGRESS(app->search_progressbar), \
				       0.0);
	}
}

/* server functions */
void add_to_replies_table(MP3 *item) {
	if (app->replies[app->nreplies] != NULL)
		mp3_free(app->replies[app->nreplies]);
	app->replies[app->nreplies] = item;
	app->nreplies++;
}

void update_search_list(MP3 *list) {
	int row;
	char *text[9];
	char buf[16];
	long speed;
	MP3 *item;
	struct timeval cur;

	if (list == NULL)
		return;

	gettimeofday(&cur, NULL);
	/*  ms */
	speed = (cur.tv_sec - app->time[list->index].tv_sec) * 1000 + \
		(cur.tv_usec - app->time[list->index].tv_usec) / 1000;
	g_snprintf(buf, sizeof (buf), "%ld ms", speed);

	while (list != NULL && app->nreplies < app->maxreplies) {
		text[0] = basename_for_windows(list->path);
		text[1] = list->title;
		text[2] = list->artist;
		text[3] = list->cmt;
		text[4] = list->size;
		text[5] = list->bitrate;
		text[6] = list->time;
		text[7] = list->user;
		text[8] = buf;
		row = gtk_clist_append(GTK_CLIST(app->search_clist), text);
		gtk_clist_set_row_data(GTK_CLIST(app->search_clist), row, list);

		list->speed = speed;
		add_to_replies_table(list);
		list = list->next;
	}

	while (list != NULL) {
		item = list;
		list = list->next;
		mp3_free(item);
	}

	/*
	if (app->nreplies >= app->maxreplies)
		app->stop_search = TRUE;
	*/
}

void update_query_list(Query *query) {
	if (app->show_query_list && query->text[0] != NULL) {
		static gint n = 0;

		if (n > 50) {
			guint len;
			len = gtk_text_get_length(GTK_TEXT(app->query_text));
			gtk_text_set_point(GTK_TEXT(app->query_text), len);
			gtk_text_backward_delete(GTK_TEXT(app->query_text), \
						len / 2);
			n = 0;
		}
		gtk_text_set_point(GTK_TEXT(app->query_text), 0);
		gtk_text_insert(GTK_TEXT(app->query_text), NULL, NULL, NULL, \
				query->text[0], strlen(query->text[0]));
		gtk_text_insert(GTK_TEXT(app->query_text), NULL, NULL, NULL, \
				"\n", 1);
		n++;
	}
}

void show_result(MP3 *list) {
	printf("Request\n");
	while (list) {
		printf("\t%s\n", list->path);
		list = list->next;
	}
}

struct sendto_data *sendto_data_new(struct sockaddr_in *addr, int len, Packet *packet) {
	struct sendto_data *data;

	data = (struct sendto_data *)xmalloc(sizeof(struct sendto_data));
	data->addr = (struct sockaddr *)xmalloc(len);
	data->len = len;
	memcpy(data->addr, addr, len);
	data->packet = packet;

	return data;
}

void sendto_data_free(struct sendto_data *data) {
	g_free(data->addr);
	g_free(data);
}

gboolean send_reply(GIOChannel *source, GIOCondition condition, gpointer data) {
	int fd, n;
	struct sendto_data *obj;
	Packet *packet;

	obj = (struct sendto_data *)data;

	fd = g_io_channel_unix_get_fd(source);

	n = sendto(fd, obj->packet->data, obj->packet->len, 0, \
		   obj->addr, obj->len);
	if (obj->packet->next != NULL) {
		packet = obj->packet;
		obj->packet = obj->packet->next;
		reply_packet_free(packet);
		g_io_add_watch(app->listen_channel, G_IO_OUT,&send_reply,data);
	} else  {
		reply_packet_free(obj->packet);
		sendto_data_free(obj);
	}

	/* event source removed */
	return FALSE;
}

gboolean on_receive_data(GIOChannel *source, GIOCondition condition, gpointer data) {
        int n, len;
        int fd;
        char addr[24];
        struct sockaddr_in caddr;

	if (app->listen_channel != source) {
		g_print("listen event source removed\n");
		/* remove event source */
		return FALSE;
	}

	fd = g_io_channel_unix_get_fd(source);

	len = sizeof(caddr);
	n = recvfrom(fd, msgbuf, sizeof(msgbuf) - 1, 0, \
		     (struct sockaddr *)&caddr, &len);

	if (n < 0)
		return TRUE;

	msgbuf[n] = '\0';
	if (is_query(msgbuf)) {
		MP3 *result;
		Query *query;

		query = query_new(app->pl, msgbuf);
		update_query_list(query);

		if (app->user_limit > 0) {
			result = mp3db_list_search(app->mp3_list, query->text);
			if (result != NULL) {
				struct sendto_data *data;
				Packet *packet;

				packet = reply_packet_new(app->pl, query, 
							  result);
				data = sendto_data_new(&caddr, sizeof(caddr), \
						       packet);
				g_io_add_watch(app->listen_channel, \
					       G_IO_OUT, \
					       &send_reply, \
					       data);
			}
		}
		query_free(query);
	} else {
		if (!app->stop_search) {
			MP3 *list;

			inet_ntop(AF_INET, &caddr.sin_addr, addr, sizeof(addr));
			list = parse_reply(app->pl, msgbuf, addr);
			update_search_list(list);
			show_status(_("Search result %d from %d users"), \
				    app->nreplies, app->nonlines);
		}
	}

        return TRUE;
}

void *user_limit_error_thread(void *data) {
	int fd = (int)data;
	char buf[2048];
	MTP *mtp;

	mtp = mtp_new(fd);

	read(fd, buf, sizeof(buf));
	mtp_send_user_limit(app->pl, mtp, buf, sizeof(buf));
	close(fd);

	mtp_free(mtp);

	return NULL;
}

int add_to_upload_clist(Upitem *uitem) {
	int row;
	char *text[4];

	text[0] = g_basename(uitem->path);
	text[1] = uitem->mtp->username;
	text[2] = "0.00";
	text[3] = "0.00";

	row = gtk_clist_prepend(GTK_CLIST(app->upload_clist), text);
	gtk_clist_set_row_data(GTK_CLIST(app->upload_clist), row, uitem);

	return row;
}

void remove_from_upload_clist(Upitem *uitem) {
	int row;

	row = gtk_clist_find_row_from_data(GTK_CLIST(app->upload_clist), (gpointer) uitem);
	gtk_clist_remove(GTK_CLIST(app->upload_clist), row);
}

/*   ð ̸ usec Ѵ */
long diff_timeval(struct timeval *a, struct timeval *b) {
	return ((a->tv_sec - b->tv_sec) * 1000000 + (a->tv_usec - b->tv_usec));
}

void check_bandwidth_limit(int sended, struct timeval *start, int speed) {
	struct timeval cur;
	long expected, measured, diff;

	gettimeofday(&cur, NULL);

	expected = (long)(sended / speed / 1024.0 * 1.0E6); /* speed kbytes/s  */
	measured = diff_timeval(&cur, start);
	if (expected > measured) {
		/*  ӵ  ε Ǵ Ȳ̴
		 * ð ؼ  ðŭ  */
		diff = expected - measured;
		if (diff > 1.0E7) {
			/* 10 ̻ ׳ 10ʸ  */
			cur.tv_sec = 10;
			cur.tv_usec = 0;
		} else {
			cur.tv_sec = diff / 1000000;
			cur.tv_usec = diff % 1000000;
		}
		/* select timeout   */
		select(0, NULL, NULL, NULL, &cur);
	}
}

void *upload_thread(void *data) {
	int len, n, sended, buflen, totol_len, send_len, ret;
	int fd;
	unsigned char buf[2048];
	fd_set rset, wset;
	FILE *fp;
	struct timeval timeout = { 30, 0 };
	struct timeval time_start;
	MP3db *item;
	Upitem *uitem = (Upitem *)data;

	buflen = sizeof(buf);
	fd = uitem->mtp->fd;

	/*  û  */
	mtp_recv_request(app->pl, uitem->mtp, buf, buflen);
	if (uitem->mtp->code != MTP_SUCCESS) {
		mtp_send_return_code(app->pl, uitem->mtp, buf, buflen);
		close(fd);
		uitem->flag = UL_ERROR;
		goto done;
	}

	/*  ߿ ãƺ */
	/* FIX ME! this code is not thread safe */
	item = mp3db_list_search_by_filename(app->mp3_list, uitem->mtp->filename);
	if (item == NULL) {
		uitem->mtp->code = MTP_FILE_NOT_FOUND;
		mtp_send_return_code(app->pl, uitem->mtp, buf, buflen);
		close(fd);
		uitem->flag = UL_ERROR;
		goto done;
	}

	fp = fopen(item->realpath, "r");
	if (fp == NULL) {
		uitem->mtp->code = MTP_FILE_NOT_FOUND;
		mtp_send_return_code(app->pl, uitem->mtp, buf, buflen);
		close(fd);
		uitem->flag = UL_ERROR;
		goto done;
	}

	if (uitem->mtp->position)
		fseek(fp, uitem->mtp->position, SEEK_SET);

	uitem->path = g_strdup(item->realpath);
	uitem->mtp->size = get_file_size(item->realpath);
	uitem->mtp->code = MTP_SUCCESS;
	mtp_send_return_code(app->pl, uitem->mtp, buf, buflen);

	uitem->flag = UL_START;

	sended = uitem->mtp->position;
	send_len = 0;
	totol_len = uitem->mtp->size - uitem->mtp->position;

	gettimeofday(&time_start, NULL);
	FD_ZERO(&rset);
	FD_ZERO(&wset);
	while (!feof(fp)) {
		FD_SET(fd, &rset);
		FD_SET(fd, &wset);
		/*  linux select
		 * struct timeval  ٲ۴ٰ Ѵ*/
		timeout.tv_sec = 30;
		timeout.tv_usec = 0;
		ret = select(fd + 1, &rset, &wset, NULL, &timeout);
		if (ret == 0)	/* timeout */
			continue;
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			else
				break;
		}

		if (FD_ISSET(fd, &rset)) {
			if (mtp_recv_progress(app->pl, uitem->mtp, buf, buflen) < 0)
				break;
		}
		if (FD_ISSET(fd, &wset)) {
			len = fread(buf, 1, buflen, fp);
			n = writen(fd, buf, len);
			if (n <= 0)
				break;
			sended += n;
			send_len += n;
		}
		if (app->quit_upload)
			break;
		if (app->use_bandwidth_limit) {
			check_bandwidth_limit(send_len, &time_start, \
					      app->bandwidth_limit);
		}
		if (app->quit_upload)
			break;
	}

	shutdown(fd, SHUT_WR);
	if (sended == uitem->mtp->size) {
		FD_SET(fd, &rset);
		timeout.tv_sec = 30;
		timeout.tv_usec = 0;
		select(FD_SETSIZE, &rset, NULL, NULL, &timeout);
		if (FD_ISSET(fd, &rset))
			n = mtp_recv_done(app->pl, uitem->mtp, buf, buflen);
		close(fd);
		uitem->flag = UL_COMPLETE;
	} else {
		close(fd);
		uitem->flag = UL_ERROR;
	}

done:
	pthread_mutex_lock(&nusermutex);
	app->nconnected--;
	if (app->quit_upload)
		pthread_cond_signal(&quit_cond);
	pthread_mutex_unlock(&nusermutex);

	return NULL;
}

void update_upload_on_start(Upitem *uitem) {
	gint row;
	char buf[32];

	row = gtk_clist_find_row_from_data(GTK_CLIST(app->upload_clist),
					   uitem);
	if (row < 0)
		row = add_to_upload_clist(uitem);
	g_snprintf(buf, sizeof(buf), "%0.2f", uitem->mtp->speed);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 2, buf);
	g_snprintf(buf, sizeof(buf), "%0.2f", uitem->mtp->progress);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 3, buf);
}

void update_upload_on_complete(Upitem *uitem) {
	gint row;
	char buf[32];

	row = gtk_clist_find_row_from_data(GTK_CLIST(app->upload_clist),
					   uitem);
	g_snprintf(buf, sizeof(buf), "%0.2f", uitem->mtp->speed);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 2, buf);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 3, \
			   _("Completed"));
	//remove_from_upload_clist(uitem);

	//mtp_free(uitem->mtp);
	//upload_item_destroy(uitem);
}

void update_upload_on_error(Upitem *uitem) {
	gint row;
	char buf[32];

	row = gtk_clist_find_row_from_data(GTK_CLIST(app->upload_clist),
					   uitem);
	g_snprintf(buf, sizeof(buf), "%0.2f", uitem->mtp->speed);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 2, buf);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 3, \
			   _("Incompleted"));
	//remove_from_upload_clist(uitem);

	//mtp_free(uitem->mtp);
	//upload_item_destroy(uitem);
}

void update_upload_on_default(Upitem *uitem) {
	gint row;
	char buf[32];

	row = gtk_clist_find_row_from_data(GTK_CLIST(app->upload_clist),
					   uitem);
	g_snprintf(buf, sizeof(buf), "%0.2f", uitem->mtp->speed);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 2, buf);
	g_snprintf(buf, sizeof(buf),
		 "%0.2f (Unknown Error)", uitem->mtp->progress);
	gtk_clist_set_text(GTK_CLIST(app->upload_clist), row, 3, buf);
}

gint update_upload(gpointer data) {
	Upitem *uitem = (Upitem *)data;

	switch (uitem->flag) {
	case UL_START:
		update_upload_on_start(uitem);
		return TRUE;
	case UL_COMPLETE:
		update_upload_on_complete(uitem);
		return FALSE;
	case UL_ERROR:
		update_upload_on_error(uitem);
		return FALSE;
	case UL_READY:
		return TRUE;
	default:
		update_upload_on_default(uitem);
		return FALSE;
	}

	/* we can never reach here */
	return FALSE;
}

gboolean on_receive_request(GIOChannel *source, GIOCondition condition, gpointer data) {
	int listenfd, connfd;
	int len;
	struct sockaddr_in caddr;
	pthread_t tid;

	if (app->service_channel != source) {
		g_print("request event source removed\n");
		/* remove event source */
		return FALSE;
	}

	listenfd = g_io_channel_unix_get_fd(source);
	len = sizeof(caddr);
	connfd = accept(listenfd, (struct sockaddr *)&caddr, &len);
	if (connfd < 0) {
		perror("accept");
		return TRUE;
	}

	if (app->nconnected < app->user_limit) {
		MTP *mtp;
		Upitem *uitem;

		uitem = upload_item_new();
		mtp = mtp_new(connfd);
		uitem->mtp = mtp;
		uitem->flag = UL_READY;

		pthread_mutex_lock(&nusermutex);
		app->nconnected++;
		pthread_mutex_unlock(&nusermutex);

		gtk_timeout_add(500, &update_upload, uitem);
		pthread_create(&tid, NULL, &upload_thread, (void *)uitem);
		pthread_detach(tid);
	} else {
		/* user limit error */
		pthread_create(&tid, NULL, &user_limit_error_thread, \
			       (void *)connfd);
		pthread_detach(tid);
	}

	return TRUE;
}

Messageitem *message_item_new(Message *msg) {
	Messageitem *mitem = (Messageitem *)xmalloc(sizeof(Messageitem));

	mitem->stat = MESSAGE_NEW;
	mitem->time = time(NULL);
	mitem->msg = msg;

	return mitem;
}

void message_item_free(Messageitem *mitem) {
	message_free(mitem->msg);
	g_free(mitem);
}

Messageitem *add_to_message_list(Message *msg) {
	Messageitem *mitem;

	mitem = message_item_new(msg);
	app->messages = g_list_append(app->messages, mitem);
	return mitem;
}

void remove_from_message_list(Messageitem *mitem) {
	app->messages = g_list_remove(app->messages, mitem);
	message_item_free(mitem);
}

gboolean update_messaging(gpointer data) {
	Messageitem *mitem = (Messageitem *)data;

	if (mitem->stat == MESSAGE_ERROR) {
		alert(_("Can't send message"));
		message_item_free(mitem);
		return FALSE;
	} else if (mitem->stat == MESSAGE_SENDED) {
		alert(_("Sent message successfully"));
		message_item_free(mitem);
		return FALSE;
	}
	return TRUE;
}

void *message_thread(void *data) {
	int n;
	Messageitem *mitem = (Messageitem *)data;

	mitem->msg->fd = connect_by_address(mitem->msg->addr, mitem->msg->port);
	if (mitem->msg->fd < 0) {
		mitem->stat = MESSAGE_ERROR;
		return NULL;
	}

	n = message_send(app->pl, mitem->msg);
	if (n <= 0) {
		close(mitem->msg->fd);
		mitem->stat = MESSAGE_ERROR;
		return NULL;
	}

	n = message_recv_got(app->pl, mitem->msg);
	if (n <= 0) {
		close(mitem->msg->fd);
		mitem->stat = MESSAGE_ERROR;
		return NULL;
	}

	close(mitem->msg->fd);
	mitem->stat = MESSAGE_SENDED;

	return NULL;
}

void start_message_send(Messageitem *mitem) {
	pthread_t tid;

	mitem->stat = MESSAGE_INPROGRESS;
	gtk_timeout_add(500, &update_messaging, mitem);
	pthread_create(&tid, NULL, &message_thread, mitem);
	pthread_detach(tid);
}

void set_message_text(GtkWidget *widget, Messageitem *mitem) {
	char buf[64];
	gchar *header;
	GdkColor fg = { 0, 0x4848, 0x3D3D, 0x8B8B };

	if (mitem == NULL)
		return;

	strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", \
		 localtime(&mitem->time));
	header = g_strdup_printf(_("Who: %s\nWhen: %s\n\n"), \
				mitem->msg->username, buf);
	/*
	header = g_strdup_printf(_("Who: %s\nWhen: %s\nWhere: %s:%d\n\n"), \
				mitem->msg->username, buf, \
				mitem->msg->addr, mitem->msg->port);
	*/

	gtk_text_freeze(GTK_TEXT(widget));
	gtk_text_backward_delete(GTK_TEXT(widget), \
				 gtk_text_get_length(GTK_TEXT(widget)));
	gtk_text_set_point(GTK_TEXT(widget), 0);
	gtk_text_insert(GTK_TEXT(widget), NULL, &fg, NULL, header, -1);
	gtk_text_insert(GTK_TEXT(widget), NULL, NULL, NULL, \
			mitem->msg->text, -1);
	gtk_text_thaw(GTK_TEXT(widget));
	g_free(header);
}

void pop_message_window(Messageitem *mitem) {
	if (app->message_window == NULL) {
		app->message_window = create_message_window(mitem);
		gtk_widget_show(app->message_window);
	} else if (mitem != NULL) {
		int row;
		char *text[3];
		char buf[64];

		strftime(buf, sizeof(buf), "%H:%M:%S", localtime(&mitem->time));
		text[0] = NULL;
		text[1] = mitem->msg->username;
		text[2] = buf;
		row = gtk_clist_append(GTK_CLIST(app->message_clist), text);
		gtk_clist_set_row_data(GTK_CLIST(app->message_clist), row, mitem);
		gtk_clist_select_row(GTK_CLIST(app->message_clist), row, 0);
		set_message_text(app->message_text, mitem);
	}
}

gboolean read_message_end(GIOChannel *source, GIOCondition condition, gpointer data) {
	Message *msg = (Message *)data;

	if (message_send_got(app->pl, msg) < 0) {
		g_io_channel_close(source);
		g_io_channel_unref(source);
		return FALSE;
	}
	g_io_channel_close(source);
	g_io_channel_unref(source);

	pop_message_window(add_to_message_list(msg));
	if (app->use_sound_on_message)
		sound_event(SOUND_ON_MESSAGE);

	return FALSE;
}

gboolean read_message(GIOChannel *source, GIOCondition condition, gpointer data) {
	Message *msg = message_new();

	msg->fd = g_io_channel_unix_get_fd(source);
	if (message_recv(app->pl, msg) < 0) {
		g_io_channel_close(source);
		g_io_channel_unref(source);
		return FALSE;
	}
	g_io_add_watch(source, G_IO_OUT, &read_message_end, msg);

	return FALSE;
}

gboolean on_receive_message(GIOChannel *source, GIOCondition condition, gpointer data) {
	int listenfd, connfd;
	int len;
	struct sockaddr_in caddr;
	GIOChannel *channel;

	if (app->message_channel != source) {
		g_print("message event source removed\n");
		/* remove event source */
		return FALSE;
	}

	listenfd = g_io_channel_unix_get_fd(source);
	len = sizeof(caddr);
	connfd = accept(listenfd, (struct sockaddr *)&caddr, &len);
	if (connfd < 0) {
		perror("accept");
		return TRUE;
	}

	channel = g_io_channel_unix_new(connfd);
	len = g_io_add_watch(channel, G_IO_IN, &read_message, NULL);

	return TRUE;
}

void start_server_process(void) {
	/* Ҹٴ listenfd io channel  */
	/* query ÿ */
	open_listen_fd(app->pl);
	app->listen_channel = g_io_channel_unix_new(get_listen_fd(app->pl));
	g_io_add_watch(app->listen_channel, G_IO_IN, &on_receive_data, NULL);

	/*  ۿ */
	app->nconnected = 0;
	open_service_fd(app->pl);
	app->service_channel = g_io_channel_unix_new(get_service_fd(app->pl));
	g_io_add_watch(app->service_channel, G_IO_IN, &on_receive_request, NULL);

	/* message  */
	open_message_fd(app->pl);
	app->message_channel = g_io_channel_unix_new(get_message_fd(app->pl));
	g_io_add_watch(app->message_channel, G_IO_IN, &on_receive_message, NULL);
}

void stop_server_process(void) {
	if (app->listen_channel) {
		g_io_channel_close(app->listen_channel);
		g_io_channel_unref(app->listen_channel); /* for init */
	}
	app->listen_channel = NULL;

	if (app->service_channel) {
		g_io_channel_close(app->service_channel);
		g_io_channel_unref(app->service_channel); /* for init */
	}
	app->service_channel = NULL;

	if (app->message_channel) {
		g_io_channel_close(app->message_channel);
		g_io_channel_unref(app->message_channel); /* for init */
	}
	app->message_channel = NULL;
}

/* login  Լ */
gint update_login_state(gpointer data) {
	int state = *((int *)data);

	switch (state) {
	case 0:
		app->quit_upload = FALSE;
		app->is_login = TRUE;

		start_server_process();

		gtk_widget_set_sensitive (app->menu_disconnect, TRUE);
		gtk_widget_set_sensitive (app->search_button, TRUE);
		gtk_widget_set_sensitive (app->search_entry, TRUE);
		gtk_widget_grab_focus(app->search_entry);
		show_status(_("connected to soribada"));
		return FALSE;
	case 1:
		show_status(_("Get address of login server"));
		return TRUE;
	case -1:
		show_status(_("Fail to get address of login server"));
		gtk_widget_set_sensitive (app->menu_connect_to_server, TRUE);
		return FALSE;
	case 2:
		show_status(_("Get login data"));
		return TRUE;
	case -2:
		show_status(_("Fail to get login data"));
		gtk_widget_set_sensitive (app->menu_connect_to_server, TRUE);
		return FALSE;
	case 3:
		show_status(_("Check local ip address"));
		return TRUE;
	case -3:
		show_status(_("No response for local ip address"));
		gtk_widget_set_sensitive (app->menu_connect_to_server, TRUE);
		return FALSE;
	case 4:
	default:
		show_status(_("Login process ..."));
		return TRUE;
	}
}

void *login_process(void *data) {
	int *state = (int *)data;

	*state = 1;
	if (get_login_url(app->pl) < 0) {
		*state = -1;
		return NULL;
	}
	*state = 2;
	if (get_login_data(app->pl) != 0) {
		*state = -2;
		return NULL;
	}
	*state = 3;
	if (app->check_online)
		if (send_local_ip(app->pl) < 0) {
			*state = -3;
			return NULL;
		}
	*state = 0;
	return NULL;
}

void login(char *username, char *password) {
	pthread_t tid;
	static int state;

	set_user_data(app->pl, username, password);
	if (app->use_proxy)
		setup_proxy_setting(app->proxy_server);

	gtk_widget_set_sensitive (app->menu_connect_to_server, FALSE);
	gtk_widget_set_sensitive (app->search_button, FALSE);
	gtk_widget_set_sensitive (app->search_entry, FALSE);

	state = 4;
	gtk_timeout_add(50, &update_login_state, (void *)&state);
	pthread_create(&tid, NULL, login_process, (void *)&state);
	pthread_detach(tid);
}

void play_mp3_file(GList *list) {
	pid_t pid;
	
	pid = fork();
	if (pid == 0) {
		/* child, play mp3 */
		int i;
		char **argv;
		char *p;

		close_listen_fd(app->pl);
		close_service_fd(app->pl);

		i = 0;
		p = strtok(app->mp3player, " ");
		while (p != NULL) {
			list = g_list_insert(list, p, i++);
			p = strtok(NULL, " ");
		}

		argv = (char **)xmalloc(sizeof(char *) * (g_list_length(list) \
					+ 1));

		for (i = 0; list != NULL; i++) {
			argv[i] = (char *)list->data;
			list = list->next;
		}
		argv[i] = NULL;

		if (execvp(argv[0], argv) < 0) {
			perror(app->mp3player);
			_exit(1);
		}
	} else if (pid > 0) {
		/* parent */
		app->mp3playerpid = pid;
	} else {
		/* error */
		perror("fork");
	}
}

void stop_mp3_playing(void) {
	if (app->mp3playerpid > 0) {
		kill(app->mp3playerpid, SIGTERM);
		app->mp3playerpid = -1;
	}
}

void wait_for_upload_thread(void) {
	/*  ε thread Ҷ ٸ */
	app->quit_upload = TRUE;
	pthread_mutex_lock(&nusermutex);
	while (app->nconnected > 0) {
		pthread_cond_wait(&quit_cond, &nusermutex);
		g_print("Main: recvd a signal\n");
	}
	pthread_mutex_unlock(&nusermutex);
	pthread_cond_init(&quit_cond, NULL);
	app->quit_upload = FALSE;
}

void quit_soribada(void) {
	if (app->is_login) {
		wait_for_upload_thread();
		stop_server_process();
		reset_search();
	}
	app->is_login = FALSE;

	sori_protocol_free(app->pl);
	app->pl = NULL;

	write_conf_file();
	write_mp3db_list();
}

Preferences *create_preferences(void) {
	GList *list;
	Preferences *obj;

	obj = (Preferences *)xmalloc(sizeof(Preferences));

	obj->lock = FALSE;
	obj->username = g_strdup(app->username);
	obj->password = g_strdup(app->password);
	obj->soribada_server = g_strdup(app->soribada_server);
	obj->proxy_server = g_strdup(app->proxy_server);
	obj->download_dir = g_strdup(app->download_dir);
	obj->mp3player = g_strdup(app->mp3player);

	obj->localmp3_dir = NULL;
	list = app->localmp3_dir;
	while (list != NULL) {
		obj->localmp3_dir = g_list_append(obj->localmp3_dir, \
						g_strdup((gchar *)list->data));
		list = list->next;
	}

	obj->port = app->port;
	obj->user_limit = app->user_limit;
	obj->maxreplies = app->maxreplies;
	obj->bandwidth_limit = app->bandwidth_limit;

	obj->width = app->width;
	obj->height = app->height;
	
	obj->auto_login = app->auto_login;
	obj->save_username = app->save_username;
	obj->save_password = app->save_password;
	obj->use_default_server = app->use_default_server;
	obj->use_bandwidth_limit = app->use_bandwidth_limit;
	obj->check_online = app->check_online;
	obj->use_proxy = app->use_proxy;
	obj->show_query_list = app->show_query_list;

	obj->sound_device = app->sound_device;
	obj->sound_on_search = g_strdup(app->sound_on_search);
	obj->sound_on_down = g_strdup(app->sound_on_down);
	obj->sound_on_message = g_strdup(app->sound_on_message);
	obj->use_sound_on_search = app->use_sound_on_search;
	obj->use_sound_on_down = app->use_sound_on_down;
	obj->use_sound_on_message = app->use_sound_on_message;

	obj->search_clist_filename = app->search_clist_filename;
	obj->search_clist_title = app->search_clist_title;
	obj->search_clist_artist = app->search_clist_artist;
	obj->search_clist_comment = app->search_clist_comment;
	obj->search_clist_size = app->search_clist_size;
	obj->search_clist_bitrate = app->search_clist_bitrate;
	obj->search_clist_time = app->search_clist_time;
	obj->search_clist_user = app->search_clist_user;
	obj->search_clist_speed = app->search_clist_speed;

	obj->downloading_clist_filename = app->downloading_clist_filename;
	obj->downloading_clist_user = app->downloading_clist_user;
	obj->downloading_clist_size = app->downloading_clist_size;
	obj->downloading_clist_received = app->downloading_clist_received;
	obj->downloading_clist_progress = app->downloading_clist_progress;
	obj->downloading_clist_speed = app->downloading_clist_speed;

	obj->downloaded_clist_filename = app->downloaded_clist_filename;
	obj->downloaded_clist_user = app->downloaded_clist_user;
	obj->downloaded_clist_size = app->downloaded_clist_size;
	obj->downloaded_clist_received = app->downloaded_clist_received;
	obj->downloaded_clist_status = app->downloaded_clist_status;
	obj->scan_sub_dirs = app->scan_sub_dirs;
	
	return obj;
}

void destroy_preferences(Preferences *obj) {
	if (obj != NULL) {
		GList *list;

		g_free(obj->username);
		g_free(obj->password);
		g_free(obj->soribada_server);
		g_free(obj->download_dir);
		g_free(obj->mp3player);
		for (list = obj->localmp3_dir; list != NULL; list = list->next)
		       g_free(list->data);	
		g_list_free(obj->localmp3_dir);

		g_free(obj->sound_on_search);
		g_free(obj->sound_on_down);
		g_free(obj->sound_on_message);

		g_free(obj);
	}
}

void copy_preferences_to_app(Preferences *obj) {
	GList *list;

	g_free(app->username);
	app->username = g_strdup(obj->username);
	g_free(app->password);
	app->password = g_strdup(obj->password);
	g_free(app->soribada_server);
	app->soribada_server = g_strdup(obj->soribada_server);
	g_free(app->proxy_server);
	app->proxy_server = g_strdup(obj->proxy_server);
	g_free(app->download_dir);
	app->download_dir = g_strdup(obj->download_dir);
	g_free(app->mp3player);
	app->mp3player = g_strdup(obj->mp3player);

	/* localmp3_dir  ʰ ׳  ű */
	for (list = app->localmp3_dir; list != NULL; list = list->next)
		       g_free(list->data);	
	g_list_free(app->localmp3_dir);
	app->localmp3_dir = obj->localmp3_dir;
	obj->localmp3_dir = NULL;

	app->port = obj->port;
	app->user_limit = obj->user_limit;
	app->maxreplies = obj->maxreplies;
	app->bandwidth_limit = obj->bandwidth_limit;

	app->width = obj->width;
	app->height = obj->height;
	app->auto_login = obj->auto_login;
	app->save_username = obj->save_username;
	app->save_password = obj->save_password;
	app->use_default_server = obj->use_default_server;
	app->use_bandwidth_limit = obj->use_bandwidth_limit;
	app->check_online = obj->check_online;
	app->use_proxy = obj->use_proxy;

	app->sound_device = obj->sound_device;
	g_free(app->sound_on_search);
	app->sound_on_search = g_strdup(obj->sound_on_search);
	g_free(app->sound_on_down);
	app->sound_on_down = g_strdup(obj->sound_on_down);
	g_free(app->sound_on_message);
	app->sound_on_message = g_strdup(obj->sound_on_message);
	app->use_sound_on_search = obj->use_sound_on_search;
	app->use_sound_on_down = obj->use_sound_on_down;
	app->use_sound_on_message = obj->use_sound_on_message;

	app->show_query_list = obj->show_query_list;
	app->search_clist_filename = obj->search_clist_filename;
	app->search_clist_title = obj->search_clist_title;
	app->search_clist_artist = obj->search_clist_artist;
	app->search_clist_comment = obj->search_clist_comment;
	app->search_clist_size = obj->search_clist_size;
	app->search_clist_bitrate = obj->search_clist_bitrate;
	app->search_clist_time = obj->search_clist_time;
	app->search_clist_user = obj->search_clist_user;
	app->search_clist_speed = obj->search_clist_speed;
	app->downloading_clist_filename = obj->downloading_clist_filename;
	app->downloading_clist_user = obj->downloading_clist_user;
	app->downloading_clist_size = obj->downloading_clist_size;
	app->downloading_clist_received = obj->downloading_clist_received;
	app->downloading_clist_progress = obj->downloading_clist_progress;
	app->downloading_clist_speed = obj->downloading_clist_speed;
	app->downloaded_clist_filename = obj->downloaded_clist_filename;
	app->downloaded_clist_user = obj->downloaded_clist_user;
	app->downloaded_clist_size = obj->downloaded_clist_size;
	app->downloaded_clist_received = obj->downloaded_clist_received;
	app->downloaded_clist_status = obj->downloaded_clist_status;
	app->scan_sub_dirs = obj->scan_sub_dirs;
}

void setup_proxy_setting(const char *addr) {
	char *addrstr;
	char *server;
	char *p;
	int port;

	if (addr == NULL)
		return;

	addrstr = g_strdup(addr);
	server = addrstr;
	p = strchr(addrstr, ':');
	if (p == NULL) {
		g_free(addrstr);
		return;
	}
	*p++ = '\0';
	if (p == '\0') {
		g_free(addrstr);
		return;
	}
	port = strtol(p, NULL, 10);

	set_proxy_use(app->pl, app->use_proxy);
	set_proxy_server(app->pl, server, port);
	g_free(addrstr);
}

/* preference ִ ˻  ϸ replies ̺ ٽ  */
void reallocate_replies_table(int newsize) {
	int i;
	MP3 **replies;

	if (newsize == app->maxreplies)
		return;

	replies = (MP3 **)xmalloc(sizeof(MP3 *) * newsize);
	for (i = 0; i < newsize; i++)
		replies[i] = NULL;

	if (newsize < app->nreplies) {
		int row;
		for (i = 0; i < newsize; i++) {
			replies[i] = app->replies[i];
			app->replies[i] = NULL;
		}

		/* search clist   */
		gtk_clist_freeze(GTK_CLIST(app->search_clist));
		for (; i < app->nreplies; i++) {
			if (app->replies[i] != NULL) {
				row = gtk_clist_find_row_from_data(GTK_CLIST(app->search_clist), app->replies[i]);
				gtk_clist_remove(GTK_CLIST(app->search_clist), row);
				mp3_free(app->replies[i]);
			}
		}
		gtk_clist_thaw(GTK_CLIST(app->search_clist));
		app->nreplies = newsize;

		for (; i < app->maxreplies; i++)
			mp3_free(app->replies[i]);

		show_status(_("Search result %d from %d users"), \
			    app->nreplies, app->nonlines);
	} else if (newsize >= app->nreplies) {
		for (i = 0; i < app->nreplies; i++) {
			replies[i] = app->replies[i];
			app->replies[i] = NULL;
		}
		for (; i < app->maxreplies; i++)
			mp3_free(app->replies[i]);
	}

	g_free(app->replies);
	app->replies = replies;
}

void do_auto_login(void) {
	if (app->auto_login && app->username && app->password)
		login(app->username, app->password);
}

File_info *file_info_new(void) {
	File_info *info = (File_info *)xmalloc(sizeof(File_info));

	info->mp3db = NULL;
	info->ditem = NULL;

	return info;
}

void file_info_free(File_info *info) {
	g_free(info);
}

/* ̹ ϴ  ˻ Ѵ, abs path
 * ٿεص Ǹ TRUE, ȵǸ FALSE */
gboolean check_filename_for_download(gchar *abspath) {
	gchar *tailed_filename;
	Downitem *ditem;

	if (abspath == NULL)
		return FALSE;

	/* ٿε   ༮ ִ ˻ */
	ditem = find_in_downloads_table_by_path(abspath);
	if (ditem != NULL) {
		alert(_("%s\n is already exist in downloading list\n" \
			"Select another file name"), \
		      g_basename(abspath));
		return FALSE;
	}

	/* ̹ ϴ  ˻ */
	if (is_file_exist(abspath)) {
		alert(_("%s\n is already exist\n" \
			"Use \"Download as ...\" menu"), \
		      g_basename(abspath));
		return FALSE;
	}

	/* tail޸ ༮ ִ  ˻ */
	tailed_filename = g_strconcat(abspath, DOWNLOAD_TAIL, NULL);
	if (is_file_exist(tailed_filename)) {
		alert(_("%s\n is already on downloading\n" \
			"Use \"Download as ...\" menu"), \
		      g_basename(abspath));
		g_free(tailed_filename);
		return FALSE;
	}
	g_free(tailed_filename);

	return TRUE;
}

void search_in_result(gchar *query, gchar *iquery) {
	int i;
	gchar *str, *p;
	gchar *keyword, *ikeyword;
	gchar *text;

	str = g_strdup(query);
	p = strtok(str, " ");
 	if (p != NULL)
		keyword = g_strdup(p);
	else
		keyword = NULL;
	g_free(str);

	str = g_strdup(iquery);
	p = strtok(str, " ");
 	if (p != NULL)
		ikeyword = g_strdup(p);
	else
		ikeyword = NULL;
	g_free(str);

	gtk_clist_freeze(GTK_CLIST(app->search_clist));
	if (keyword != NULL && strlen(keyword) > 0) {
		for (i = 0; i < GTK_CLIST(app->search_clist)->rows; i++) {
			gtk_clist_get_text(GTK_CLIST(app->search_clist),
					   i, 0, &text);
			if (xstrstr(text, keyword) != NULL)
				continue;
			gtk_clist_get_text(GTK_CLIST(app->search_clist),
					   i, 1, &text);
			if (xstrstr(text, keyword) != NULL)
				continue;
			gtk_clist_get_text(GTK_CLIST(app->search_clist),
					   i, 2, &text);
			if (xstrstr(text, keyword) != NULL)
				continue;
			gtk_clist_remove(GTK_CLIST(app->search_clist), i);
			i--;
		}
	}
	if (ikeyword != NULL && strlen(ikeyword) > 0) {
		for (i = 0; i < GTK_CLIST(app->search_clist)->rows; i++) {
			gtk_clist_get_text(GTK_CLIST(app->search_clist),
					   i, 0, &text);
			if (xstrstr(text, ikeyword) != NULL) {
				gtk_clist_remove(GTK_CLIST(app->search_clist), i);
				i--;
				continue;
			}
			gtk_clist_get_text(GTK_CLIST(app->search_clist),
					   i, 1, &text);
			if (xstrstr(text, ikeyword) != NULL) {
				gtk_clist_remove(GTK_CLIST(app->search_clist), i);
				i--;
				continue;
			}
			gtk_clist_get_text(GTK_CLIST(app->search_clist),
					   i, 2, &text);
			if (xstrstr(text, ikeyword) != NULL) {
				gtk_clist_remove(GTK_CLIST(app->search_clist), i);
				i--;
				continue;
			}
		}
	}
	g_free(keyword);
	g_free(ikeyword);
	gtk_clist_thaw(GTK_CLIST(app->search_clist));
	app->redraw_search_clist = TRUE;
}

void redraw_search_clist(void) {
	int i;
	int row;
	char *text[8];
	char buf[16];
	MP3 *item;

	gtk_clist_freeze(GTK_CLIST(app->search_clist));
	gtk_clist_clear(GTK_CLIST(app->search_clist));
	for (i = 0; i < app->nreplies; i++) {
		item = app->replies[i];
		if (item != NULL) {
			g_snprintf(buf, sizeof(buf), "%d ms", item->speed);
			text[0] = basename_for_windows(item->path);
			text[1] = item->title;
			text[2] = item->artist;
			text[3] = item->size;
			text[4] = item->bitrate;
			text[5] = item->time;
			text[6] = item->user;
			text[7] = buf;

			row = gtk_clist_append(GTK_CLIST(app->search_clist), text);
			gtk_clist_set_row_data(GTK_CLIST(app->search_clist), row, item);
		}
	}
	gtk_clist_thaw(GTK_CLIST(app->search_clist));
	show_status(_("Search result %d from %d users"), \
		    app->nreplies, app->nonlines);
}

