/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __CORE_H_
#define __CORE_H_

#include "soribada.h"

typedef struct mp3db_item {
	char *realpath;
	MP3 *data;
} MP3db;

struct sendto_data {
	struct sockaddr *addr;
	int len;
	Packet *packet; 
};

typedef struct file_info_s {
	MP3db *mp3db;
	Downitem *ditem;
	gchar *path;
	GtkWidget *filename;

	gboolean use_id3_tag;
	GtkWidget *title;
	GtkWidget *artist;
	GtkWidget *album;
	GtkWidget *comment;
	GtkWidget *year;
	GtkWidget *track;
	GtkWidget *genre;
	gint genre_no;

	GtkWidget *save_button;
	GtkWidget *rename_button;
} File_info;

typedef struct message_item_s {
	enum {
		MESSAGE_NEW,
		MESSAGE_READ,

		MESSAGE_INPROGRESS,
		MESSAGE_ERROR,
		MESSAGE_SENDED
	} stat;
	time_t time;
	Message *msg;
} Messageitem;

typedef enum {
	SOUND_ON_SEARCH,
	SOUND_ON_DOWN,
	SOUND_ON_MESSAGE
} Sound_Event;

void show_status(const char *fmt, ...);

void login(char *username, char *password);
void do_auto_login(void);
void stop_server_process(void);
void wait_for_upload_thread(void);
void quit_soribada(void);
void setup_proxy_setting(const char *addr);

void reallocate_replies_table(int newsize);
void remove_from_replies_table(struct mp3_item *item);

Downitem *download_item_new(struct mp3_item *item);
void download_item_destroy(Downitem *ditem);
void download_item_ref(Downitem *ditem);
void download_item_unref(Downitem *ditem);

Downitem *add_to_downloads_table(Downitem *item);
Downitem *add_to_download_queue(Downitem *ditem);
Downitem *remove_from_downloads_table(Downitem *ditem);
Downitem *remove_from_download_queue(Downitem *ditem);
Downitem *find_in_downloads_table_by_path(gchar *path);

int add_to_downloading_clist(Downitem *item);

/* download */
gboolean check_filename_for_download(gchar *abspath);
void start_download(Downitem *item);

/* search */
void search_in_result(gchar *query, gchar *iquery);
void redraw_search_clist(void);
void start_search(const gchar *query_string);
gboolean stop_search(gpointer data);
void reset_search(void);

void play_mp3_file(GList *list);
void stop_mp3_playing(void);

/* preferences */
Preferences *create_preferences(void);
void destroy_preferences(Preferences *obj);
void copy_preferences_to_app(Preferences *obj);

/* file info */
File_info *file_info_new(void);
void file_info_free(File_info *info);

/* message window */
Messageitem *message_item_new(Message *msg);
void message_item_free(Messageitem *mitem);
void set_message_text(GtkWidget *widget, Messageitem *mitem);
void pop_message_window(Messageitem *mitem);
void start_message_send(Messageitem *mitem);

/* mp3list.c */
gboolean is_mp3(const char *name);
GList *load_mp3db_list(void);
void write_mp3db_list(void);
void delete_mp3db_list(void);
void scan_local_mp3_dir(void);

void mp3db_item_free(MP3db *item);
GList *mp3db_list_append(GList *list, char *mp3file);
GList *mp3db_list_load_from_file(char *listfile);
void mp3db_list_write_to_file(GList *list, char *filename);
MP3 *mp3db_list_search(GList *list, char **text);
MP3db *mp3db_list_search_by_filename(GList *list, char *filename);
MP3db *mp3db_list_search_by_realpath(GList *list, char *filename);

/* sound.c */
void sound_event(Sound_Event event);
void play_wave_file(const gchar *filename, Sound_Device device);

#endif /* __CORE_H_ */
