/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>

#include <gtk/gtk.h>

void *xmalloc(size_t size) {
	void *item;

	item = malloc(size);
	if (item == NULL) {
		fprintf(stderr, "Memory allocation error\n");
		exit(1);
	}

	return item;
}

char *basename_for_windows(char *path) {
	char *p = strchr(path, '\0');

	while (p > path && *p != '\\')
		p--;

	if (*p == '\\')
		return p + 1;

	return p;
}

gboolean is_file_exist(const char *filename) {
	return (access(filename, F_OK) == 0);
}

int get_file_size(const char *filename) {
	struct stat st;

	stat(filename, &st);

	return st.st_size;
}

gint compare_by_num(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2) {
	char *text1 = NULL;
	char *text2 = NULL;
	GtkCListRow *row1 = (GtkCListRow *)ptr1;
	GtkCListRow *row2 = (GtkCListRow *)ptr2;

	text1 = GTK_CELL_TEXT(row1->cell[clist->sort_column])->text;
	text2 = GTK_CELL_TEXT(row2->cell[clist->sort_column])->text;

	return strtol(text1, NULL, 10) - strtol(text2, NULL, 10);
}

int writen(int fd, const void *vptr, int n) {
	int nleft, nwritten;
	const char *ptr;

	ptr = (char *)vptr;
	nleft = n;
	while (nleft > 0) {
		if ((nwritten = write(fd, ptr, nleft)) <= 0) {
			if (errno == EINTR)
				nwritten = 0;
			else
				return -1;
		}
		nleft -= nwritten;
		ptr += nwritten;
	}
	return n - nleft;
}

void remove_crlf_from_str(char *str) {
	if (str != NULL) {
		while (*str != '\0') {
			if (*str == '\r')
				*str = ' ';
			if (*str == '\n')
				*str = ' ';
			str++;
		}
	}
}

/*  Լ strcasestr ϱ   ̴.
 *  strcasestr GNU extention Լ ⺻ POSIX  ִ Լ
 * ƴϴ. ׷ٺ װ  configure.in strcasestr üũϴ 
 * ũθ ߰ؾϰ  Ǵ  Ȯϰ GNU_SOURCE 
 * define ϴµ   autoconf ͼ   ׽Ʈ
 * ƾ ۼ ʰ glibc-2.2/sysdeps/generic/strcasestr.c ״
 * ͼ ԵǾ.  Լ  Ӹӿ Ŷ 
 * ʱ ٶ. ^^
 * Ϻ Լ  κ  ĳҴ.
 * Ʒ ڰ ޾Ƴ ̴ּ */
/*
 * My personal strstr() implementation that beats most other algorithms.
 * Until someone tells me otherwise, I assume that this is the
 * fastest implementation of strstr() in C.
 * I deliberately chose not to comment it.  You should have at least
 * as much fun trying to understand it, as I had to write it :-).
 *
 * Stephen R. van den Berg, berg@pool.informatik.rwth-aachen.de	*/
char *xstrstr(const char *phaystack, const char *pneedle) {
	register const unsigned char *haystack, *needle;
	register int b, c;

	haystack = (const unsigned char *) phaystack;
	needle = (const unsigned char *) pneedle;

	b = tolower (*needle);
	if (b != '\0') {
		haystack--;			/* possible ANSI violation */
		do {
			c = *++haystack;
			if (c == '\0')
				goto ret0;
		} while (tolower (c) != b);

		c = tolower (*++needle);
		if (c == '\0')
			goto foundneedle;
		++needle;
		goto jin;

		for ( ; ; ) {
			register int a;
			register const unsigned char *rhaystack, *rneedle;

			do {
				a = *++haystack;
				if (a == '\0')
					goto ret0;
				if (tolower (a) == b)
					break;
				a = *++haystack;
				if (a == '\0')
					goto ret0;
				shloop:
			} while (tolower (a) != b);

			jin:	  a = *++haystack;
			if (a == '\0')
				goto ret0;

			if (tolower (a) != c)
				goto shloop;

			rhaystack = haystack-- + 1;
			rneedle = needle;
			a = tolower (*rneedle);

			if (tolower (*rhaystack) == a)
				do {
					if (a == '\0')
						goto foundneedle;
					++rhaystack;
					a = tolower (*++needle);
					if (tolower (*rhaystack) != a)
						break;
					if (a == '\0')
						goto foundneedle;
					++rhaystack;
					a = tolower (*++needle);
				} while (tolower (*rhaystack) == a);

			needle = rneedle;	/* took the register-poor approach */

			if (a == '\0')
				break;
		}
	}

	foundneedle:
	return (char*) haystack;

	ret0:
	return 0;
}
