/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>

#include "soribada.h"
#include "mp3.h"
#include "core.h"

gboolean is_mp3(const char *name) {
        char *ep;

	ep = strchr(name, '\0');

        if (ep == NULL)
                return FALSE;
        
	if (ep - name > 4) {
		ep -= 4;
		if ( ep[0] == '.' && \
		    (ep[1] == 'm' || ep[1] == 'M') && \
		    (ep[2] == 'p' || ep[2] == 'P') && \
		     ep[3] == '3')
		return TRUE;
	}

        return FALSE;
}

MP3db *mp3db_item_new(void) {
	MP3db *item = (MP3db *)xmalloc(sizeof(MP3db));

	item->realpath = NULL;
	item->data = mp3_new();

	return item;
}

MP3db *mp3db_item_new_by_mp3(char *mp3) {
	int size;
	MP3db *item;
	struct mp3info info;

	if (mp3_get_info(mp3, &info) < 0)
		return NULL;

	remove_crlf_from_str(info.artist);
	remove_crlf_from_str(info.title);
	remove_crlf_from_str(info.comment);
	size = get_file_size(mp3);

	item = (MP3db *)xmalloc(sizeof(MP3db));
	item->realpath = strdup(mp3);

	item->data = mp3_new();
	item->data->path =  g_strdup_printf("D:\\%s", g_basename(mp3));
	item->data->size = g_strdup_printf("%d", size);
	item->data->time = g_strdup_printf("%02d:%02d", \
				     info.seconds / 60, info.seconds % 60);
	if (info.bitrate < 0)
		item->data->bitrate = g_strdup("VBR");
	else
		item->data->bitrate = g_strdup_printf("%d", info.bitrate);
	item->data->artist = g_strdup_printf("%s", info.artist);
	item->data->title = g_strdup_printf("%s", info.title);
	item->data->cmt = g_strdup_printf("%s", info.comment);
	item->data->next = NULL;

	return item;
}

void mp3db_item_free(MP3db *item) {
	g_free(item->realpath);
	mp3_free(item->data);
	g_free(item);
}

GList *mp3db_list_load_from_file(char *listfile) {
	FILE *fp;

	fp = fopen(listfile, "r");
	if (fp != NULL) {
		char buf[1024];
		GList *list = NULL;
		MP3db *item;

		while (fgets(buf, sizeof(buf), fp) != NULL) {
			item = mp3db_item_new();
			item->realpath = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->path = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->size = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->time = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->bitrate = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->artist = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->title = g_strdup(g_strchomp(buf));
			fgets(buf, sizeof(buf), fp);
			item->data->cmt = g_strdup(g_strchomp(buf));

			list = g_list_append(list, (gpointer)item);
		}

		fclose(fp);
		return list;
	} else {
		perror(listfile); 
		return NULL;
	}
}

GList *mp3db_list_load(void) {
	char *listfile;
	GList *list;

	listfile = g_strdup_printf("%s/%s", getenv("HOME"), MP3_LIST);
	if (access(listfile, R_OK) == 0) {
		list = mp3db_list_load_from_file(listfile);
		g_free(listfile);
		return list;
	}
	g_free(listfile);

	return NULL;
}

GList *mp3db_list_append(GList *list, char *mp3file) {
	MP3db *item;

	item = mp3db_item_new_by_mp3(mp3file);
	if (item == NULL)
		return NULL;

	return g_list_append(list, (gpointer)item);
}

GList *mp3db_list_remove(GList *list, MP3db *item) {
	return g_list_remove(list, (gpointer)item);
}

gboolean test_keyword(char *text, char **keyword) {
	while (*keyword != NULL) {
		if (xstrstr(text, *keyword) == NULL)
			return FALSE;
		keyword++;
	}
	return TRUE;
}

MP3 *mp3db_list_search(GList *list, char **text) {
	char **pp;
	MP3 *item;
	MP3 *result = NULL;

	while (list != NULL) {
		pp = text;
		item = ((MP3db *)(list->data))->data;
		//g_print("name: %s\n", item->name);

		if (test_keyword(item->path, text)) {
			result = mp3_list_append(result, item);
			continue;
		}

		if (test_keyword(item->artist, text)) {
			result = mp3_list_append(result, item);
			continue;
		}

		if (test_keyword(item->title, text)) {
			result = mp3_list_append(result, item);
			continue;
		}

		if (test_keyword(item->cmt, text)) {
			result = mp3_list_append(result, item);
			continue;
		}

		list = list->next;
	}
	return result;
}

MP3db *mp3db_list_search_by_filename(GList *list, char *filename) {
	MP3 *item;

	while (list != NULL) {
		item = ((MP3db *)list->data)->data;
		if (strcmp(item->path, filename) == 0)
			return (MP3db *)list->data;
		list = list->next;
	}
	return NULL;
}

MP3db *mp3db_list_search_by_realpath(GList *list, char *filename) {
	MP3db *item;

	while (list != NULL) {
		item = (MP3db *)list->data;
		if (strcmp(item->realpath, filename) == 0)
			return item;
		list = list->next;
	}
	return NULL;
}

GList *mp3db_list_new_by_scan(GList *list, char *dirname, gboolean subscan) {
	char *parent;
	char *cwd;
	char path[1024];
	struct stat st;
	struct dirent *entry;
	MP3db *item;
	DIR *dir;

	parent = g_get_current_dir();
	dir = opendir(dirname);
	if (dir == NULL) {
		perror(dirname);
		return NULL;
	}

	chdir(dirname);
	cwd = g_get_current_dir();
	while ((entry = readdir(dir)) != NULL) {
		if (strcmp(entry->d_name, ".") == 0)
			continue;
		if (strcmp(entry->d_name, "..") == 0)
			continue;

		if (subscan) {
			stat(entry->d_name, &st);
			if (S_ISDIR(st.st_mode))
				list = mp3db_list_new_by_scan(list,
							      entry->d_name,
							      subscan);
		}

		if (is_mp3(entry->d_name)) {
			g_snprintf(path, sizeof(path), "%s/%s", cwd, entry->d_name);
			item = mp3db_item_new_by_mp3(path);
			if (item != NULL)
				list = g_list_append(list, (gpointer)item);
		}
	}
	g_free(cwd);

	chdir(parent);
	closedir(dir);
	g_free(parent);

	return list;
}

static void free_mp3db_item(gpointer data, gpointer user_data) {
	mp3db_item_free((MP3db *)data);
}

void mp3db_list_free(GList *list) {
	g_list_foreach(list, &free_mp3db_item, NULL);
	g_list_free(list);
}

void write_data(gpointer data, gpointer user_data) {
	fputs(((MP3db *)data)->realpath, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->path, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->size, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->time, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->bitrate, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->artist, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->title, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
	fputs(((MP3db *)data)->data->cmt, (FILE *)user_data);
	fputc('\n', (FILE *)user_data);
}

void mp3db_list_write_to_file(GList *list, char *filename) {
	FILE *fp;

	fp = fopen(filename, "w");
	if (fp != NULL) {
		g_list_foreach(list, &write_data, (gpointer)fp);
		fclose(fp);
	} else {
		perror(filename); 
	}
}

void write_mp3db_list(void) {
	mp3db_list_write_to_file(app->mp3_list, app->mp3_list_file);
}

void delete_mp3db_list(void) {
	if(app->mp3_list != NULL)
		mp3db_list_free(app->mp3_list);
	app->mp3_list = NULL;
}

void scan_local_mp3_dir(void) {
	GList *list;

	delete_mp3db_list();
	list = app->localmp3_dir;
	while (list != NULL) {
		app->mp3_list = mp3db_list_new_by_scan(app->mp3_list, \
						       (char *)list->data, \
						       app->scan_sub_dirs);
		list = list->next;
	}
}

