/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* protocol.c
 *   ̺귯  ϱ ؼ glib ִ ƿ
 *  ʾҽϴ.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "protocol.h"

/* Native Language Support. */
#ifdef ENABLE_NLS
#include <libintl.h>
#include <locale.h>
#define _(String) gettext(String)
#else 
#define _(String) String
#endif
#define N_(String) String

#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

#define return_if_null(arg)	do {\
					if (arg == NULL) return; \
				} while (0)

#define free_if_not_null(ptr) 	do {\
					if (ptr) free(ptr); \
				} while (0)

static void *xmalloc(int size) {
	void *ret = malloc(size);

	if (ret == NULL) {
		fprintf(stderr, _("Memory allocation error\n"));
		exit(1);
	}

	return ret;
}

/* strtokʹ  ٸ
 * ־ str pattern  ڸ.  ̶ pattern شϴ ڵ
 * nulڰ Ǿ . reentrancy  cur_ptr  */
static char *tokener(char *str, char *pattern, char **cur_ptr) {
	char *p, *sp;
	int i;

	if (str)
		*cur_ptr = str;

	if (*cur_ptr == NULL)
		return NULL;

	sp = *cur_ptr;
	p = sp;
	i = 0;
	while (*p != '\0') {
		if (pattern[i] == '\0')
			break;

		if (*p == pattern[i])
			i++;
		else {
			p -= i;
			i = 0;
		}
		p++;
	}

	if (*p == '\0') {
		if (pattern[i] == '\0') {
			for (; i > 0; i--) {
				*--p = '\0';
			}
		}
		*cur_ptr = NULL;
	} else {
		*cur_ptr = p;
		for (; i > 0; i--) {
			*--p = '\0';
		}
	}

	return sp;
}

static int readn(int fd, void *vptr, int n) {
	int nleft, nread;
	char *ptr;

	ptr = vptr;
	nleft = n;
	while (nleft > 0) {
		if ((nread = read(fd, ptr, nleft)) < 0) {
			if (errno == EINTR)
				nread = 0;
			else
				return -1;
		} else if (nread == 0)
			break;
		nleft -= nread;
		ptr += nread;
	}

	return (n - nleft);
}

static int writen(int fd, const void *vptr, int n) {
	int nleft, nwritten;
	const char *ptr;

	ptr = (char *)vptr;
	nleft = n;
	while (nleft > 0) {
		if ((nwritten = write(fd, ptr, nleft)) <= 0) {
			if (errno == EINTR)
				nwritten = 0;
			else
				return -1;
		}
		nleft -= nwritten;
		ptr += nwritten;
	}
	return n;
}

char *get_hostname_from_url(const char *url) {
	if (url != NULL && \
	    url[0] == 'h' && url[1] == 't' && url[2] == 't' && url[3] == 'p' &&\
	    url[4] == ':' && url[5] == '/' && url[6] == '/') {
		int i;
		char hostname[256];

		/* url  ȣƮ   */
		for (i = 0; url[i + 7] != '\0' && url[i + 7] != '/'; i++)
			hostname[i] = url[i + 7];

		hostname[i] = '\0';

		return strdup(hostname);
	}
	return NULL;
}

char *get_querystr_from_url(const char *url) {
	if (url != NULL && \
	    url[0] == 'h' && url[1] == 't' && url[2] == 't' && url[3] == 'p' &&\
	    url[4] == ':' && url[5] == '/' && url[6] == '/') {
		const char *p;

		/* url  query string  */
		p = url + 7;
		while (*p != '\0' && *p != '/')
			p++;

		return strdup(p);
	}
	return NULL;
}

int connect_by_hostname(const char *hostname, int port) {
	int sockfd;
	struct sockaddr_in servaddr;
	struct hostent *hent;

	hent = gethostbyname(hostname);
	if (hent == NULL) {
		herror("gethostbyname");
		return -1;
	}

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd == -1) {
		perror("socket");
		return -1;
	}

	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(port);
	memcpy(&servaddr.sin_addr, *(struct in_addr **)hent->h_addr_list,
	       sizeof(struct in_addr));

	if (connect(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) != 0) {
		perror("connect");
		return -1;
	}

	return sockfd;
}

int connect_by_address(const char *addr, int port) {
	int sockfd;
	struct sockaddr_in servaddr;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd == -1) {
		perror("socket");
		return -1;
	}

	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(port);
	inet_pton(AF_INET, addr, &servaddr.sin_addr);

	if (connect(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) != 0) {
		perror("connect");
		return -1;
	}

	return sockfd;
}

int connect_by_address_nb(const char *addr, int port, int sec) {
	int fd;
	int flags, ret, error, len;
	fd_set rset, wset;
	struct timeval timeout;
	struct sockaddr_in servaddr;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd == -1) {
		perror("socket");
		return -1;
	}

	/* Nonblocking I/O  ȯ */
	flags = fcntl(fd, F_GETFL, 0);
	if (flags < 0) {
		perror("fcntl F_GETFL");
		close(fd);
		return -1;
	}
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0) {
		perror("fcntl F_SETFL");
		close(fd);
		return -1;
	}

	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(port);
	inet_pton(AF_INET, addr, &servaddr.sin_addr);

	error = 0;
	ret = connect(fd, (struct sockaddr *)&servaddr, sizeof(servaddr));
	/* ret == 0 ̸ ٷ ӵ ̴ */
	if (ret < 0) {
		if (errno != EINPROGRESS) {
			close(fd);
			return -1;
		}

		/* select ̿ؼ Ѵ */
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_SET(fd, &rset);
		FD_SET(fd, &wset);
		timeout.tv_sec = sec;
		timeout.tv_usec = 0;

		ret = select(fd + 1, &rset, &wset, NULL, &timeout);
		if (ret == 0) {
			close(fd);
			errno = ETIMEDOUT;
			return -1;
		}
		if (FD_ISSET(fd, &rset) || FD_ISSET(fd, &wset)) {
			len = sizeof (error);
			if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) <0) {
				/* solaris error */
				perror("connect");
				close(fd);
				return -1;
			}
		}
	}

	/* blocking I/O  ȯ */
	flags &= ~O_NONBLOCK; 		/*  Non Blocking */
	if (fcntl(fd, F_SETFL, flags) < 0) {
		perror("F_SETFL error");
		close(fd);
		return -1;
	}

	if (error) {
		/* BSD error */
		close(fd);
		errno = error;
		perror("connect");
		return -1;
	}

	return fd;
}

static unsigned char * send_query_by_url(const char *url) {
	char *hostname, *querystr, *ret, *answer, *p;
	char query[1024];
	unsigned char *buf[256];
	int sockfd, port, len, n = 0, i, j;

	hostname = get_hostname_from_url(url);
	querystr = get_querystr_from_url(url);

	if (hostname == NULL || querystr == NULL) {
		free_if_not_null(hostname);
		free_if_not_null(querystr);
		return NULL;
	}

	len = snprintf(query, sizeof (query),
			"GET %s HTTP/1.0\r\n"
			"Host: %s\r\n"
			"Accept: www/source, text/html, video/mpeg, image/jpeg, image/x-tiff\r\n"
			"Accept: image/x-rgb, image/x-xbm, image/gif, */*, application/postscript\r\n"
			"Content-type: application/x-www-form-urlencoded\r\n\r\n",
			querystr, hostname);

	port = 80;
	sockfd = connect_by_hostname(hostname, port);
	if (sockfd < 0) {
		free_if_not_null(hostname);
		free_if_not_null(querystr);
		return NULL;
	}

	writen(sockfd, query, len);

	for (i = 0; i < 256; i++)
		buf[i] = NULL;
	len = 0;
	for (i = 0; i < 256; i++) {
		buf[i] = (char *)xmalloc(8192);
		n = readn(sockfd, buf[i], 8192);
		len += n;
		if (n < 8192)
			break;
	}

	close(sockfd);

	answer = (char *)xmalloc(len + 1);
	p = answer;
	for (j = 0; j < i; j++) {
		memcpy(p, buf[j], 8192);
		free(buf[j]);
		p += 8192;
	}
	memcpy(p, buf[j], n);
	free(buf[j]);
	p[n] = '\0';

	free(hostname);
	free(querystr);

	p = strstr(answer, "\r\n\r\n");
	if (p == NULL) {
		free(answer);
		return NULL;
	} else {
		ret = strdup(p + 4);
		free(answer);
		return ret;
	}
}

static unsigned char *
send_query_by_url_proxy(const char *url, const char *proxy, int port) {
	char *hostname, *querystr, *ret, *answer, *p;
	char query[1024];
	unsigned char *buf[256];
	int sockfd, len, n = 0, i, j;

	hostname = get_hostname_from_url(url);
	querystr = get_querystr_from_url(url);

	if (hostname == NULL || querystr == NULL) {
		free_if_not_null(hostname);
		free_if_not_null(querystr);
		return NULL;
	}

	len = snprintf(query, sizeof (query),
			"GET http://%s%s HTTP/1.0\r\n"
			"Host: %s\r\n"
			"Accept: www/source, text/html, video/mpeg, image/jpeg, image/x-tiff\r\n"
			"Accept: image/x-rgb, image/x-xbm, image/gif, */*, application/postscript\r\n"
			"Content-type: application/x-www-form-urlencoded\r\n\r\n",
			hostname, querystr, hostname);

	sockfd = connect_by_hostname(proxy, port);
	if (sockfd < 0) {
		free_if_not_null(hostname);
		free_if_not_null(querystr);
		return NULL;
	}

	writen(sockfd, query, len);

	for (i = 0; i < 256; i++)
		buf[i] = NULL;
	len = 0;
	for (i = 0; i < 256; i++) {
		buf[i] = (char *)xmalloc(8192);
		n = readn(sockfd, buf[i], 8192);
		len += n;
		if (n < 8192)
			break;
	}

	close(sockfd);

	answer = (char *)xmalloc(len + 1);
	p = answer;
	for (j = 0; j < i; j++) {
		memcpy(p, buf[j], 8192);
		free(buf[j]);
		p += 8192;
	}
	memcpy(p, buf[j], n);
	free(buf[j]);
	p[n] = '\0';

	free(hostname);
	free(querystr);

	p = strstr(answer, "\r\n\r\n");
	if (p == NULL) {
		free(answer);
		return NULL;
	} else {
		ret = strdup(p + 4);
		free(answer);
		return ret;
	}
}

/* On Success : return 0, else return -1 */
int get_login_url(Protocol *pl) {
	char *login_server;
	char *answer;
	char url[1024];

	snprintf(url, sizeof (url), "http://%s/gateway.txt", pl->server);

	if (pl->use_proxy)
		answer = send_query_by_url_proxy(url, pl->proxy, pl->proxy_port);
	else
		answer = send_query_by_url(url);

	if (answer == NULL)
		return -1;

	login_server = strstr(answer, "IS=");
	if (login_server == NULL) {
		return -1;
	}

	free_if_not_null(pl->login_url);
	pl->login_url = strdup(login_server + 3);
	free(answer);

	return 0;
}

/* dotted quad  ּҸ 4Ʈ array  ش */
unsigned char *make_ip_arr_from_str(char *ip_str) {
	int a, b, c, d;
	unsigned char *ip_arr = (unsigned char *)xmalloc(4);

	sscanf(ip_str, "%d.%d.%d.%d", &a, &b, &c, &d);
	ip_arr[0] = (unsigned char)a;
	ip_arr[1] = (unsigned char)b;
	ip_arr[2] = (unsigned char)c;
	ip_arr[3] = (unsigned char)d;

	return ip_arr;
}

/*  ִ  ĽϿ Protocol  εѴ */
/* on success: return 0 else return error code */
int parse_login_data(Protocol *pl, char *buf) {
	char *p, *lp;

	p = strtok_r(buf, "\r\n", &lp);
	while (p) {
		if (strncmp("Version : ", p, 10) == 0) {
			free_if_not_null(pl->version);
			pl->version = strdup(p + 10);
		} else if (strcmp("ADDON BEGIN", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			while (p != NULL) {
				if (strcmp("ADDON END", p) == 0)
					break;
				p = strtok_r(NULL, "\r\n", &lp);
			}
		} else if (strcmp("ROOMTYPE BEGIN", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			while (p != NULL) {
				if (strcmp("ROOMTYPE END", p) == 0)
					break;
				p = strtok_r(NULL, "\r\n", &lp);
			}
		} else if (strcmp("ASR", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			free_if_not_null(pl->service_url);
			pl->service_url = strdup(p);
		} else if (strcmp("ACH", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			free_if_not_null(pl->chatting_url);
			pl->chatting_url = strdup(p);
		} else if (strcmp("ACO", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strcmp("MYTM", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strcmp("MYIP", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			free_if_not_null(pl->local_ip_str);
			pl->local_ip_str = strdup(p);
			free_if_not_null(pl->local_ip);
			pl->local_ip = make_ip_arr_from_str(p);
		} else if (strcmp("PI", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strcmp("CI", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strcmp("COVER", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strcmp("COVERZIP", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strcmp("UPGRADE", p) == 0) {
			p = strtok_r(NULL, "\r\n", &lp);
			p = strtok_r(NULL, "\r\n", &lp);
			p = strtok_r(NULL, "\r\n", &lp);
		} else if (strncmp("Return Code : ", p, 14) == 0) {
			/* this is really error */
			return -1;
		} else {
			/* 𸣴 Ű Ѵ */
			printf("unknown: %s\n", p);
		}
		p = strtok_r(NULL, "\r\n", &lp);
	}

	if (pl->service_url == NULL || pl->local_ip == NULL)
		return -1;

	return 0;
}

int get_login_data(Protocol *pl) {
	int ret;
	char *answer;
	char url[1024];

	snprintf(url, sizeof (url),
		 "%s?username=%s&password=%s&version=1.94",
		 pl->login_url, pl->username, pl->password);

	if (pl->use_proxy)
		answer = send_query_by_url_proxy(url, pl->proxy, pl->proxy_port);
	else
		answer = send_query_by_url(url);

	if (answer == NULL)
		return -1;

	ret = parse_login_data(pl, answer);
	free(answer);

	return ret;
}

static int get_magic_number(int flag) {
	static int magic_number[4] = {
		0xA5,		/* 0 */
		0xA5,		/* 4 */
		0xC3,		/* 8 */
		0x77		/* C */
	};

	return magic_number[(flag & 0xF) / 4];
}

static int get_decoded_ip(int flag, int quad, int magic) {
	if (flag) {
		switch (quad) {
		case 0x16:
			return 0xA5;
		case 0x17:
	                return 0xC3;
	        case 0x18:
	                return 0x77;
	        default:
			printf("Except value(decoder): 0x%02X\n", quad);
	                return 0xC3;
		}
	} else
		return quad ^ magic;
}

static int get_encoded_ip(int no, int *flag, int quad, int magic) {
	if (quad == magic) {
		*flag = *flag | no;
		switch (quad) {
		case 0xA5:
			return 0x16;
		case 0xC3:
			return 0x17;
		case 0x77:
			return 0x18;
		default:
			printf("Except value(encoder): %d\n", quad);
			return 0x17;
		}
	} else
		return quad ^ magic;
}

static void swap(int *a, int *b) {
	int tmp = *b;

	*b = *a;
	*a = tmp;
}

/* Ҹٴٿ  ·  ͸ decencoding
 * flag, port, a, b, c, d  ־  Ѵ */
static void decode_ip_addr(Protocol *pl, int flag, int *port, int *a, int *b, int *c, int *d) {
	int magic = get_magic_number(flag);

	/* port */
	*port = pl->port_base + ((((flag & 0xF) % 4) << 8) | *port);

	/* ip addr */
	flag = (flag >> 4) & 0xF;
	*a = get_decoded_ip(flag & 0x1, *a, magic);
	*b = get_decoded_ip(flag & 0x2, *b, magic);
	*c = get_decoded_ip(flag & 0x4, *c, magic);
	*d = get_decoded_ip(flag & 0x8, *d, magic);

	swap(a, b);
	swap(b, d);
}

/* Ҹٴٿ  · encoding, 
 * ̸ flag, port, a, b, c, d  ־ callѴ */
static void encode_ip_addr(Protocol *pl, int *flag, int *port, int *a, int *b, int *c, int *d) {
	int magic;

	magic = get_magic_number(*flag);
	*flag |= ((*port - pl->port_base) >> 4) & 0xF0;
	*port = (*port - pl->port_base) & 0xFF;

	swap(b, d);
	swap(a, b);

	*a = get_encoded_ip(0x10, flag, *a, magic);
	*b = get_encoded_ip(0x20, flag, *b, magic);
	*c = get_encoded_ip(0x40, flag, *c, magic);
	*d = get_encoded_ip(0x80, flag, *d, magic);
}

int send_local_ip(Protocol *pl) {
	int flag, port, ip[4];
	char *answer;
	char url[1024];

	flag = 0x4;
	port = pl->port;
	ip[0] = pl->local_ip[0];
	ip[1] = pl->local_ip[1];
	ip[2] = pl->local_ip[2];
	ip[3] = pl->local_ip[3];

	encode_ip_addr(pl, &flag, &port, &ip[0], &ip[1], &ip[2], &ip[3]);
	snprintf(url, sizeof (url),
		 "%s?action=ping&username=%s&password=%s&ip=%%%02X%%%02X%%%02X%%%02X%%%02X%%%02X",
		 pl->service_url, pl->username, pl->password,
		 port, flag, ip[0], ip[1], ip[2], ip[3]);

	if (pl->use_proxy)
		answer = send_query_by_url_proxy(url, pl->proxy, pl->proxy_port);
	else
		answer = send_query_by_url(url);

	if (answer == NULL)
		return -1;
	
	free(answer);

	return 0;
}

int parse_online_list(Protocol *pl, unsigned char *data, struct node **nodetable) {
	int i;
	int flag, port, ip[4];
	unsigned char *p;

	if (data[0] != 'V' || data[1] != 'e' || data[2] != 'r' || \
	    data[3] != 's' || data[4] != 'i' || data[5] != 'o' || \
	    data[6] != 'n' || data[7] != ' ' || data[8] != ':' || \
	    data[9] != ' ') {
		return -1;
	}

	p = data + 10;
	while (*p != '\0' && *p != '\n')
		p++;
	p++;

	for (i = 0; i < MAX_ONLINES;) {
		port = *p++;
		flag = *p++;
		if (*p == '\0')
			break;

		ip[0] = *p++;
		ip[1] = *p++;
		ip[2] = *p++;
		ip[3] = *p++;

		decode_ip_addr(pl, flag, &port, &ip[0], &ip[1], &ip[2], &ip[3]);
		if (ip[0] != (int)pl->local_ip[0] ||
		    ip[1] != (int)pl->local_ip[1] ||
		    ip[2] != (int)pl->local_ip[2] ||
		    ip[3] != (int)pl->local_ip[3]) {
			nodetable[i]->port = port;
			nodetable[i]->flag = flag;
			snprintf(nodetable[i]->addr, 20,
				 "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
			//fprintf(stdout, "%s\n", nodetable[i]->addr);
			i++;
		}
		if (*p == '\0')
			break;
	}

	return i;
}

int get_online_list(Protocol *pl, struct node **nodetable) {
	int ret;
	unsigned char *answer;
	char url[1024];

	snprintf(url, sizeof (url),
		 "%s?action=gimme&username=%s&password=%s",
		 pl->service_url, pl->username, pl->password);

	if (pl->use_proxy)
		answer = send_query_by_url_proxy(url, pl->proxy, pl->proxy_port);
	else
		answer = send_query_by_url(url);

	if (answer == NULL)
		return -1;

	ret = parse_online_list(pl, answer, nodetable);
	free(answer);

	return ret;
}

int get_online_list_from_file(Protocol *pl, char *filename, unsigned char *data, struct node **nodetable) {
	unsigned char *answer;
	struct stat buf;
	FILE *fp;

	fp = fopen(filename, "r");
	if (fp == NULL)
		return 0;

	fstat(fileno(fp), &buf);
	answer = (char *)xmalloc(buf.st_size + 1);
	fread(answer, buf.st_size, 1, fp);

	parse_online_list(pl, answer, nodetable);

	free(answer);
	fclose(fp);

	return 0;
}

Packet *query_packet_new(Protocol *pl, const char *query) {
	int flag, port, ip[4];
	int len;
	char *p, *sp;
	unsigned char pkt[256];
	char *str;
	Packet *packet;

	flag = 0x4;	/* ϴ    ϴ 𸣴ϱ ׳ 0x4  */
	port = pl->port;
	ip[0] = pl->local_ip[0];
	ip[1] = pl->local_ip[1];
	ip[2] = pl->local_ip[2];
	ip[3] = pl->local_ip[3];
	encode_ip_addr(pl, &flag, &port, &ip[0], &ip[1], &ip[2], &ip[3]);

	pkt[0] = 1;
	pkt[1] = port;
	pkt[2] = flag;

	pkt[3] = ip[0];
	pkt[4] = ip[1];
	pkt[5] = ip[2];
	pkt[6] = ip[3];

	/*
	fprintf(stdout, "pkt: %02X %02X %02X %02X %02X %02X\n",
		pkt[1], pkt[2], pkt[3], pkt[4], pkt[5], pkt[6]);
	*/

	pkt[7] = 0;
	pkt[8] = 0;

	pkt[9] = 'Q';
	pkt[10] = ':';
	pkt[11] = '\0';

	len = 11;
	str = strdup(query);
	p = tokener(str, " ", &sp);
	while (p != NULL) {
		strncat(&pkt[9], "+", sizeof(pkt) - 11);
		len++;
		strncat(&pkt[9], p, sizeof(pkt) - 11);
		len += strlen(p);
		strncat(&pkt[9], "\n", sizeof(pkt) - 11);
		len++;
		p = tokener(NULL, " ", &sp);
		if (len >= sizeof(pkt))
			break;
	}
	free(str);

	packet = (Packet *)xmalloc(sizeof(Packet));
	packet->data = (unsigned char *)xmalloc(len);
	/* pkt[7,8]  ⺻ 0̹Ƿ strcpy()   */
	memcpy(packet->data, pkt, len);
	packet->len = len;

	return packet;
}

void query_packet_free(Packet *packet) {
	return_if_null(packet);

	free_if_not_null(packet->data);
	free_if_not_null(packet);
}

void update_query_packet_id(Packet *packet, unsigned short int id) {
	packet->data[7] = (unsigned char)(id & 0xFF);
	packet->data[8] = (unsigned char)((id >> 8) & 0xFF);
}

int is_reply(unsigned char *buf) {
	return (buf != NULL && buf[9] == 'R' && buf[10] == ':');
}

MP3 *parse_reply(Protocol *pl, unsigned char *buf, char *addr) {
	if (is_reply(buf)) {
		unsigned short int index;
		char *cp;
		char *p = tokener(buf + 11, "\n", &cp);
		char *user;
		int port;
		MP3 *item = NULL;
		MP3 *list = NULL;

		index = (unsigned short)buf[8] << 8 | (unsigned short)buf[7];
		user = p;
		p = tokener(NULL, "\n", &cp);
		port = atoi(p);
		p = tokener(NULL, "\n", &cp);

		while (p != NULL) {
			item = mp3_new();
			item->index = index;
			item->user = strdup(user);
			item->addr = strdup(addr);
			item->port = port;
			item->path = strdup(p);
			p = tokener(NULL, "\n", &cp);
			item->size = strdup(p);
			p = tokener(NULL, "\n", &cp);
			item->time = strdup(p);
			p = tokener(NULL, "\n", &cp);
			item->bitrate = strdup(p);
			p = tokener(NULL, "\n", &cp);
			item->artist = strdup(p);
			p = tokener(NULL, "\n", &cp);
			item->title = strdup(p);
			p = tokener(NULL, "\n", &cp);
			item->cmt = strdup(p);
			p = tokener(NULL, "\r\n", &cp);
			item->something = strdup(p);
			p = tokener(NULL, "\n", &cp);

			list = mp3_list_append(list, item);
		}
		return list;
	} else
		return NULL;
}

int is_query(unsigned char *buf) {
	return (buf != NULL && buf[9] == 'Q' && buf[10] == ':');
}

Query *query_new(Protocol *pl, unsigned char *buf) {
	int flag, port, ip[4];
	int i;
	char *p, *sp;
	Query *item = (Query *)xmalloc(sizeof(Query));

	port = buf[1];
	flag = buf[2];
	ip[0] = buf[3];
	ip[1] = buf[4];
	ip[2] = buf[5];
	ip[3] = buf[6];

	decode_ip_addr(pl, flag, &port, &ip[0], &ip[1], &ip[2], &ip[3]);

	item->port = port;
	item->ip[0] = ip[0];
	item->ip[1] = ip[1];
	item->ip[2] = ip[2];
	item->ip[3] = ip[3];

	item->index = (buf[8] << 8) | buf[7];

	p = strtok_r(buf + 12, "\r\n+- ", &sp);
	for (i = 0; i < 7 && p != NULL; i++) {
		item->text[i] = strdup(p);
		p = strtok_r(NULL, "\r\n+- ", &sp);
	}
	for ( ; i < 8; i++)
		item->text[i] = NULL;

	return item;
}

void query_free(Query *query) {
	return_if_null(query);

	free_if_not_null(query->text[0]);
	free_if_not_null(query->text[1]);
	free_if_not_null(query->text[2]);
	free_if_not_null(query->text[3]);
	free_if_not_null(query->text[4]);
	free_if_not_null(query->text[5]);
	free_if_not_null(query->text[6]);
	free_if_not_null(query->text[7]);
	free_if_not_null(query);
}

Packet *reply_packet_new(Protocol *pl, Query *query, MP3 *list) {
	int flag, port, ip[4];
	int i, len, hdrlen, buflen = 1024;
	MP3 *item;
	Packet *cur, *next;
	Packet *packet = (Packet *)xmalloc(sizeof(Packet));

	packet->data = (unsigned char *)xmalloc(buflen);
	packet->next = NULL;

	flag = 0x4;
	port = pl->port;
	ip[0] = pl->local_ip[0];
	ip[1] = pl->local_ip[1];
	ip[2] = pl->local_ip[2];
	ip[3] = pl->local_ip[3];
	encode_ip_addr(pl, &flag, &port, &ip[0], &ip[1], &ip[2], &ip[3]);

	packet->data[0] = 1;
	packet->data[1] = port;
	packet->data[2] = flag;

	packet->data[3] = ip[0];
	packet->data[4] = ip[1];
	packet->data[5] = ip[2];
	packet->data[6] = ip[3];

	packet->data[7] = (unsigned char)(query->index & 0xFF);
	packet->data[8] = (unsigned char)(query->index >> 8);

	packet->data[9] = 'R';
	packet->data[10] = ':';

	cur = packet;
	item = list;
	i = 11;
	i += snprintf(packet->data + i, buflen - i, \
		      "%s\n%4d\n", pl->username, pl->port + 1);
	hdrlen = i;
	while (item != NULL) {
		i += snprintf(cur->data + i, buflen - i, \
			      "%s\n%s\n%s\n%s\n%s\n%s\n%s\n\r\n",
			      item->path, item->size, item->time, item->bitrate,
			      item->artist, item->title, item->cmt);
		item = item->next;
		if (item != NULL) {
			len = strlen(item->path) + 1;
			len += strlen(item->size) + 1;
			len += strlen(item->time) + 1;
			len += strlen(item->bitrate) + 1;
			len += strlen(item->artist) + 1;
			len += strlen(item->title) + 1;
			len += strlen(item->cmt) + 3;
			if (i + len > buflen) { 
				next = (Packet *)xmalloc(sizeof(Packet));
				next->data = (unsigned char *)xmalloc(buflen);
				next->next = NULL;
				memcpy(next->data, packet->data, hdrlen);

				cur->next = next;
				cur->len = i;

				cur = next;
				i = hdrlen;
			}
		} else
			cur->len = i;
	}

	return packet;
}

void reply_packet_free(Packet *packet) {
	return_if_null(packet);

	free_if_not_null(packet->data);
	free_if_not_null(packet);
}

MP3 *mp3_list_last(MP3 *list) {
	if (list != NULL) {
		while (list->next != NULL)
			list = list->next;
	}

	return list;
}

MP3 *mp3_list_append(MP3 *list, MP3 *item) {
	if (list != NULL) {
		MP3 *last;

		last = mp3_list_last(list);
		item->next = NULL;
		last->next = item;

		return list;
	} else {
		item->next = NULL;

		return item;
	}
}

void mp3_list_free(MP3 *list) {
	MP3 *item;

	while (list != NULL) {
		item = list;
		list = list->next;
		mp3_free(item);
	}
}

MTP *mtp_new(int fd) {
	socklen_t len;
	struct sockaddr_in addr;
	MTP *mtp = (MTP *)xmalloc(sizeof(MTP));

	mtp->fd = fd;
	mtp->filename = NULL;
	mtp->size = 0;
	mtp->portm = 0;
	mtp->position = 0;
	mtp->progress = 0;
	mtp->speed = 0;
	mtp->username = NULL;
	mtp->msg = NULL;

	len = sizeof(addr);
	if (getpeername(fd, (struct sockaddr *)&addr, &len) < 0) {
		perror("getpeername");
		mtp->addr = NULL;
	} else {
		char buf[32];

		if (inet_ntop(AF_INET, &addr.sin_addr, buf, sizeof(buf)))
			mtp->addr = strdup(buf);
		else
			mtp->addr = NULL;
	}

	return mtp;
}

void mtp_free(MTP *mtp) {
	return_if_null(mtp);

	free_if_not_null(mtp->filename);
	free_if_not_null(mtp->username);
	free_if_not_null(mtp->addr);
	free_if_not_null(mtp->msg);
	free_if_not_null(mtp);
}

int mtp_send_request(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	int len;

	len = snprintf(buf, buflen,
			"GETMP3\r\n"
			"Filename: %s\r\n"
			"Position: %ld\r\n"
			"PortM: %d\r\n"
			"Username: %s\r\n\r\n",
			mtp->filename,
			mtp->position,
			pl->port + 3,
			pl->username);
	return writen(mtp->fd, buf, len);
}

int mtp_send_user_limit(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	int len;

	len = snprintf(buf, buflen, "MTP 1.0 100 User Limit\r\n\r\n");

	return writen(mtp->fd, buf, len);
}

int mtp_recv_return_code(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	char *p, *cp;
	int n, size_ok, position_ok;

	n = read(mtp->fd, buf, buflen - 1);
	if (n < 0)
		return -1;
	buf[n] = '\0';

	p = buf;
	/* MTP check */
	if (p[0] != 'M' || p[1] != 'T' || p[2] != 'P')
		return n;

	/* MTP string */
	p = tokener(buf, " ", &cp);
	if (p == NULL)
		return n;

	/* version */
	p = tokener(NULL, " ", &cp);
	if (p == NULL)
		return n;

	/* return code */
	p = tokener(NULL, " ", &cp);
	if (p == NULL)
		return n;
	mtp->code = strtol(p, NULL, 10);
	
	/* return message */
	p = tokener(NULL, "\r\n", &cp);
	if (p == NULL)
		return n;
	mtp->msg = strdup(p);

	if (mtp->code != 0)
		return n;

	size_ok = FALSE;
	position_ok = FALSE;
	p = tokener(NULL, ": ", &cp);
	while (p != NULL) {
		if (strcmp(p, "Filesize") == 0) {
			/* file size */
			p = tokener(NULL, "\r\n", &cp);
			if (p == NULL)
				return -1;
			mtp->size = strtol(p, NULL, 10);
			size_ok = TRUE;
		} else if (strcmp(p, "Position") == 0) {
			/* position */
			p = tokener(NULL, "\r\n", &cp);
			if (p == NULL)
				return -1;
			mtp->position = strtol(p, NULL, 10);
			position_ok = TRUE;
		}
		p = tokener(NULL, ": ", &cp);
	}
	if (size_ok && position_ok) {
		mtp->code = MTP_SUCCESS;
		return n;
	} else 
		return 0;
}

int mtp_send_progress(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	int len;

	mtp->progress = (float) mtp->received / (float)mtp->size * 100;

	len = snprintf(buf, buflen,
		       "GOT\r\n"
		       "Progress: %.2f\r\n"
		       "Speed: %.2f\r\n\r\n",
		       mtp->progress, mtp->speed);

	return writen(mtp->fd, buf, len);
}

int mtp_send_done(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	int len;

	len = snprintf(buf, buflen,
		       "GOT\r\n"
		       "Progress: 100.00\r\n"
		       "Speed: %.2f\r\n\r\n"
		       "DONE\r\n\r\n",
		       mtp->speed);

	return writen(mtp->fd, buf, len);
}

const char *get_mtp_err_msg(int code) {
	switch (code) {
	case MTP_SUCCESS:
		return "Succeded";	/* öڸ Ʋ  ´ */
	case MTP_USER_LIMIT:
		return "User Limit";
	case MTP_FILE_NOT_FOUND:
		return "File Not Found";
	case MTP_INVALID_COMMAND:
	default:
		return "Invalid Command";
	}
}

int mtp_recv_request(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	char *p, *cp;
	int n;
	unsigned int ret = 0;

	n = read(mtp->fd, buf, buflen);
	if (n < 0) {
		mtp->code = MTP_ERROR;
		return -1;
	} else if (n == 0) {
		return -1;
	}

	buf[n] = '\0';
	p = tokener(buf, "\r\n", &cp);
	if (strcmp(p, "GETMP3") != 0) {
		mtp->code = MTP_INVALID_COMMAND;
		return 0;
	}

	p = tokener(NULL, "\r\n", &cp);
	while (p != NULL) {
		if (strncmp(p, "Filename: ", 10) == 0) {
			free_if_not_null(mtp->filename);
			mtp->filename = strdup(p + 10);
			ret |= 0x1;
		} else if (strncmp(p, "Position: ", 10) == 0) {
			mtp->position = strtol(p + 10, NULL, 10);
			ret |= 0x2;
		} else if (strncmp(p, "PortM: ", 7) == 0) {
			mtp->portm = strtol(p + 7, NULL, 10);
			ret |= 0x4;
		} else if (strncmp(p, "Username: ", 10) == 0) {
			free_if_not_null(mtp->username);
			mtp->username = strdup(p + 10);
			ret |= 0x8;
		}

		p = tokener(NULL, "\r\n", &cp);
	}

	if (ret == 0xF)
		mtp->code = MTP_SUCCESS;
	else
		mtp->code = MTP_INVALID_COMMAND;

	return 0;
}

int mtp_send_return_code(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	int len;

	len = snprintf(buf, buflen, "MTP 1.0 %03d %s\r\n", \
		     mtp->code, get_mtp_err_msg(mtp->code));

	if (mtp->code == MTP_SUCCESS) {
		len += snprintf(buf + len, buflen - len,
			     "Filesize: %d\r\nPosition: %ld\r\n",
			     mtp->size, mtp->position);
	}

	return writen(mtp->fd, buf, len);
}

int mtp_recv_progress(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	char *p, *cp;
	int n;

	n = read(mtp->fd, buf, buflen);
	if (n < 0) {
		mtp->code = MTP_ERROR;
		return -1;
	} else if (n == 0) {
		return -1;
	}
	buf[n] = '\0';

	p = tokener(buf, "\r\n", &cp);
	while (p != NULL) {
		if (strcmp(p, "GOT") == 0) {
			; 		/* do nothing */
		} else if (strncmp(p, "Progress: ", 10) == 0) {
			mtp->progress = strtod(p + 10, NULL);
		} else if (strncmp(p, "Speed: ", 7) == 0) {
			mtp->speed = strtod(p + 7, NULL);
		} else if (strcmp(p, "DONE") == 0) {
			mtp->code = MTP_DONE;
		}
		p = tokener(NULL, "\r\n", &cp);
	}

	return n;
}

int mtp_recv_done(Protocol *pl, MTP *mtp, char *buf, int buflen) {
	char *p, *cp;
	int n;

	n = read(mtp->fd, buf, buflen);
	if (n < 0) {
		mtp->code = MTP_ERROR;
		return -1;
	} else if (n == 0) {
		return 0;
	}
	buf[n] = '\0';

	p = tokener(buf, "\r\n", &cp);
	while (p != NULL) {
		if (strcmp(p, "GOT") == 0) {
			; 		/* do nothing */
		} else if (strncmp(p, "Progress: ", 10) == 0) {
			mtp->progress = strtod(p + 10, NULL);
		} else if (strncmp(p, "Speed: ", 7) == 0) {
			mtp->speed = strtod(p + 7, NULL);
		} else if (strcmp(p, "DONE") == 0) {
			mtp->code = MTP_DONE;
		}
		p = tokener(NULL, "\r\n", &cp);
	}
	return n;
}

Protocol *sori_protocol_new(const char *remote_server, int local_port) {
	Protocol *pl;

	pl = (Protocol *)xmalloc(sizeof(Protocol));

	pl->server = strdup(remote_server);
	pl->version = NULL;

	pl->username = NULL;
	pl->password = NULL;

	pl->use_proxy = FALSE;
	pl->proxy = NULL;

	pl->listenfd = -1;
	pl->servicefd = -1;
	pl->messagefd = -1;
	pl->port = local_port;
	pl->port_base = PORT_BASE;

	pl->local_ip_str = NULL;
	pl->local_ip = NULL;
	pl->service_url = NULL;
	pl->chatting_url = NULL;
	pl->login_url = NULL;

	return pl;
}

void sori_protocol_free(Protocol *pl) {
	return_if_null(pl);

	free_if_not_null(pl->server);
	free_if_not_null(pl->version);

	free_if_not_null(pl->proxy);

	free_if_not_null(pl->local_ip_str);
	free_if_not_null(pl->local_ip);
	free_if_not_null(pl->service_url);
	free_if_not_null(pl->chatting_url);
	free_if_not_null(pl->login_url);

	free_if_not_null(pl);
}

void set_user_data(Protocol *pl, char *username, char *password) {
	free_if_not_null(pl->username);
	pl->username = strdup(username);

	free_if_not_null(pl->password);
	pl->password = strdup(password);
}

void set_proxy_use(Protocol *pl, int flag) {
	pl->use_proxy = flag;
}

void set_proxy_server(Protocol *pl, const char *proxy, int port) {
	free_if_not_null(pl->proxy);
	pl->proxy_port = port;
}

int open_socket(int port, int type) {
	int fd;
	struct sockaddr_in addr;

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(port);

	fd = socket(AF_INET, type, 0);
	if (fd == -1) {
		perror("socket");
		return -1;
	}

	if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
		char buf[32];

		snprintf(buf, sizeof(buf), "bind: port: %d ", port);
		perror(buf);
		return -1;
	}

	return fd;
}

int open_listen_fd(Protocol *pl) {
	pl->listenfd = open_socket(pl->port, SOCK_DGRAM);
	return pl->listenfd;
}

int close_listen_fd(Protocol *pl) {
	return close(pl->listenfd);
}

int get_listen_fd(Protocol *pl) {
	return pl->listenfd;
}

int open_service_fd(Protocol *pl) {
	int flags;
	pl->servicefd = open_socket(pl->port + 1, SOCK_STREAM);
	if (listen(pl->servicefd, 5) < 0) {
		perror("listen");
		return -1;
	}

	/* non blocking io */
	flags = fcntl(pl->servicefd, F_GETFL, 0);
	if (flags < 0)
		perror("fcntl: F_GETFL");
	else {
		flags |= O_NONBLOCK;
		if (fcntl(pl->servicefd, F_SETFL, flags) < 0)
			perror("fcntl: F_SETFL");
	}

	return pl->servicefd;
}

int close_service_fd(Protocol *pl) {
	return close(pl->servicefd);
}

int get_service_fd(Protocol *pl) {
	return pl->servicefd;
}

int open_message_fd(Protocol *pl) {
	int flags;
	pl->messagefd = open_socket(pl->port + 3, SOCK_STREAM);
	if (listen(pl->messagefd, 5) < 0) {
		perror("listen");
		return -1;
	}

	/* non blocking io */
	flags = fcntl(pl->messagefd, F_GETFL, 0);
	if (flags < 0)
		perror("fcntl: F_GETFL");
	else {
		flags |= O_NONBLOCK;
		if (fcntl(pl->messagefd, F_SETFL, flags) < 0)
			perror("fcntl: F_SETFL");
	}

	return pl->messagefd;
}

int close_message_fd(Protocol *pl) {
	return close(pl->messagefd);
}

int get_message_fd(Protocol *pl) {
	return pl->messagefd;
}

MP3 * mp3_new(void) {
	MP3 *item = (MP3 *)xmalloc(sizeof(MP3));

	item->index = 0;
	item->user = NULL;
	item->addr = NULL;
	item->port = 0;
	item->path = NULL;
	item->size = NULL;
	item->time = NULL;
	item->bitrate = NULL;
	item->artist = NULL;
	item->title = NULL;
	item->cmt = NULL;
	item->something = NULL;
	item->speed = 0;
	item->next = NULL;

	return item;
}

void mp3_free(MP3 *item) {
	return_if_null(item);

	free_if_not_null(item->user);
	free_if_not_null(item->addr);
	free_if_not_null(item->path);
	free_if_not_null(item->size);
	free_if_not_null(item->time);
	free_if_not_null(item->bitrate);
	free_if_not_null(item->artist);
	free_if_not_null(item->title);
	free_if_not_null(item->cmt);
	free_if_not_null(item->something);
	free_if_not_null(item);
}

Message *message_new(void) {
	Message *msg = (Message *)xmalloc(sizeof(Message));

	msg->fd = -1;
	msg->username = NULL;
	msg->addr = NULL;
	msg->port = -1;
	msg->text = NULL;

	return msg;
}

Message *message_new_from_mp3(MP3 *mp3) {
	Message *msg = (Message *)xmalloc(sizeof(Message));

	msg->fd = -1;
	msg->username = strdup(mp3->user);
	msg->addr = strdup(mp3->addr);
	/* mp3->port tcp Ʈ̹Ƿ 9002  
	 * ׷ 3 ƴ 2 Ѵ */
	msg->port = mp3->port + 2; 
	msg->text = NULL;

	return msg;
}

Message *message_new_from_mtp(MTP *mtp) {
	Message *msg = (Message *)xmalloc(sizeof(Message));

	msg->fd = -1;
	msg->username = strdup(mtp->username);
	if (mtp->addr == NULL)
		msg->addr = NULL;
	else
		msg->addr = strdup(mtp->addr);
	msg->port = mtp->portm; 
	msg->text = NULL;

	return msg;
}

void message_free(Message *msg) {
	return_if_null(msg);

	free_if_not_null(msg->username);
	free_if_not_null(msg->addr);
	free_if_not_null(msg->text);
	free_if_not_null(msg);
}

int message_send(Protocol *pl, Message *msg) {
	int n, len, buflen;
	char buf[2048];

	/* buf  2048 ߱  ̺  ޼
	 * ׳ 2048 ߶  */
	buflen = sizeof(buf) - 1;
	n = snprintf(buf, buflen, \
		     "From:%s\007" \
		     "Address:%s-%d\007" \
		     "Message:", \
		     pl->username, \
		     pl->local_ip_str, \
		     pl->port + 3);
	len = strlen(msg->text);
	len = len > buflen - n ? buflen - n : len;
	memcpy(buf + n, msg->text, len);
	n += len;
	buf[n++] = 0x07;

	return writen(msg->fd, buf, n);
}

int message_send_got(Protocol *pl, Message *msg) {
	return write(msg->fd, "RC: Got", 7);
}

int message_recv(Protocol *pl, Message *msg) {
	int n;
	char *p, *sp, *sep;
	char buf[2048];

	n = read(msg->fd, buf, sizeof(buf) - 1);
	if (n <= 0)
		return n;
	buf[n] = '\0';

	/* msg  ʵ ڰ 0x07̴ */
	/* From: */
	p = buf;
	sep = strchr(p, 0x07);
	if (sep == NULL)
		return -1;
	*sep = '\0';
	if (strncmp(p, "From:", 5) != 0)
		return -1;
	msg->username = strdup(p + 5);

	/* address */
	if (sep + 1 >= &buf[n])
		return -1;
	p = sep + 1;
	sep = strchr(p, 0x07);
	if (sep == NULL)
		return -1;
	*sep = '\0';
	if (strncmp(p, "Address:", 8) != 0)
		return -1;
	p = strtok_r(p + 8, "-", &sp);
	if (p == NULL)
		return -1;
	msg->addr = strdup(p);
	msg->port = strtol(sp, NULL, 10);

	/* message text */
	if (sep + 1 >= &buf[n])
		return -1;
	p = sep + 1;
	sep = strchr(p, 0x07);
	if (sep == NULL)
		return -1;
	*sep = '\0';
	if (strncmp(p, "Message:", 8) != 0)
		return -1;
	msg->text = strdup(p + 8);

	return 0;
}

int message_recv_got(Protocol *pl, Message *msg) {
	int n;
	char buf[32];

	n = read(msg->fd, buf, sizeof(buf) - 1);
	if (n <= 0)
		return n;
	buf[n] = '\0';

	if (strcmp(buf, "RC: Got") == 0)
		return n;
	else
		return -1;
}
