/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* protocol.h
 * Ҹٴ  ̿ϰ  
 */

#ifndef __PROTOCOL_H_
#define __PROTOCOL_H_

#define PORT_BASE       9000
#define MAX_ONLINES	8000

struct node {
	char addr[20];	/* ٸ Ŭ̾Ʈ IP addr: string */
	int port;	/* ٸ Ŭ̾Ʈ port */
	int flag;	/* decoding  ѹ */
};

enum MTP_error_code {
	MTP_SUCCESS = 0,
	MTP_USER_LIMIT = 100,
	MTP_DOWNLOAD_LIMIT = 101,
	MTP_FILE_NOT_FOUND = 102,
	MTP_INVALID_COMMAND = 103,
	MTP_DONE = 1000,
	MTP_ERROR = 2000
};

typedef struct mtp_s {
	int fd;
	enum MTP_error_code code;
	char *filename;
	int size;
	int portm;
	int received;
	float speed;
	float progress;
	long position;
	char *username;
	char *addr;
	char *msg;
} MTP;

typedef struct mp3_item {
	unsigned int index;
	char *user;
	char *addr;
	int port;
	char *path;
	char *size;
	char *time;
	char *bitrate;
	char *artist;
	char *title;
	char *cmt;
	char *something;
	int speed;
	struct mp3_item *next;
} MP3; 

typedef struct query_s {
	unsigned short int index;
	int port;
	unsigned char ip[4];
	char *text[8];
} Query;

typedef struct packet_s {
	unsigned char *data;
	int len;
	struct packet_s *next;
} Packet;

typedef struct message_s {
	int fd;
	char *username;
	char *addr;
	int port;
	char *text;
} Message;

typedef struct protocol_s {
	char *server;			/*  ּ */
	char *version;			/*   */
	char *username;
	char *password;

	/* login Ҷ proxy     ɼ */
	int use_proxy;			/* proxy  ϴ° */
	char *proxy;			/* proxy  ּ */
	int proxy_port;			/* proxy  Ʈ */

	int listenfd;			/* listen socket(udp) */
	int servicefd;			/* listen socket(tcp) */
	int messagefd;			/* listen to message */
	int port;			/* listen port */
	int port_base;			/* listen port base */

	char *local_ip_str;		/*  IP addr: string */
	unsigned char *local_ip;	/*  IP addr: arrary */
	char *service_url;		/* ¶  ޴  */
	char *chatting_url;		/* ä   */
	char *login_url;		/* login ϴ  */
} Protocol;

Protocol *sori_protocol_new(const char *remote_server, int local_port);
void sori_protocol_free(Protocol *pl);
void set_user_data(Protocol *pl, char *username, char *password);
void set_proxy_use(Protocol *pl, int flag);
void set_proxy_server(Protocol *pl, const char *proxy, int port);

int open_listen_fd(Protocol *pl);
int close_listen_fd(Protocol *pl);
int get_listen_fd(Protocol *pl);
int open_service_fd(Protocol *pl);
int close_service_fd(Protocol *pl);
int get_service_fd(Protocol *pl);
int open_message_fd(Protocol *pl);
int close_message_fd(Protocol *pl);
int get_message_fd(Protocol *pl);

int connect_by_address(const char *addr, int port);
int connect_by_address_nb(const char *addr, int port, int sec);

/* login */
int get_login_url(Protocol *pl);
int get_login_data(Protocol *pl);
int send_local_ip(Protocol *pl);

/* online data */
int get_online_list(Protocol *pl, struct node **nodetable);
int get_online_list_from_file(Protocol *pl, char *filename, unsigned char *data, struct node **nodetable);

int is_query(unsigned char *buf);
int is_reply(unsigned char *buf);
MP3 *parse_reply(Protocol *pl, unsigned char *buf, char *addr);

MP3 *mp3_new(void);
void mp3_free(MP3 *item);

/* query Ÿ: 츮  query ؼ   */
Query *query_new(Protocol *pl, unsigned char *buf);
void query_free(Query *query);

MP3 *mp3_list_append(MP3 *list, MP3 *item);
void mp3_list_free(MP3 *list);

/* reply packet  */
Packet *reply_packet_new(Protocol *pl, Query *query, MP3 *list);
void reply_packet_free(Packet *packet);

/* query packet  Լ */
Packet *query_packet_new(Protocol *pl, const char *query);
void query_packet_free(Packet *packet);
void update_query_packet_id(Packet *packet, unsigned short int id);

/* mpt client  ְ ޴  */
MTP*	mtp_new(int fd);
void 	mtp_free(MTP *mtp);
/* upload */
int mtp_send_user_limit(Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_recv_request (Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_send_return_code (Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_recv_progress (Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_recv_done (Protocol *pl, MTP *mtp, char *buf, int buflen);
/* download */
int mtp_send_request (Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_recv_return_code (Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_send_progress (Protocol *pl, MTP *mtp, char *buf, int buflen);
int mtp_send_done (Protocol *pl, MTP *mtp, char *buf, int buflen);

/* message  */
Message *message_new(void);
Message *message_new_from_mp3(MP3 *mp3);
Message *message_new_from_mtp(MTP *mtp);
void message_free(Message *msg);
int message_send(Protocol *pl, Message *msg);
int message_send_got(Protocol *pl, Message *msg);
int message_recv(Protocol *pl, Message *msg);
int message_recv_got(Protocol *pl, Message *msg);

#endif /* __PROTOCOL_H_ */
