/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __SORIBADA_H_
#define __SORIBADA_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stddef.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <gtk/gtk.h>

/* Native Language Support. */
#ifdef ENABLE_NLS
#include <libintl.h>
#include <locale.h>
#define _(String) gettext(String)
#else 
#define _(String) String
#endif
#define N_(String) String

#include "protocol.h"

#define DEFAULT_SERVER	"www.soribada.com"
#define DEFAULT_PORT	9001
#define PORT_BASE	9000

#define CONF_FILE	".soribadarc"
#define MP3_LIST	".soribadalist"
#define DOWNLOAD_TAIL	".ing"
#define DEFAULT_MP3_PLAYER	"mpg123"

#define MAX_ONLINES     8000
#define MAX_TIME_TABLE	65536	
#define MAX_SEARCH_RESULT	1000
#define DEFAULT_MAX_REPLIES	500

#define DEFAULT_SOUND_SEARCH	"search.wav"
#define DEFAULT_SOUND_DOWN	"down.wav"
#define DEFAULT_SOUND_MESSAGE	"message.wav"

#ifndef TRUE
	#define TRUE		1
#endif
#ifndef FALSE
	#define FALSE		0
#endif

typedef struct download_item {
	enum {
		DL_CMD_NONE = 1,
		DL_CMD_RUN,
		DL_CMD_STOP,
		DL_CMD_REMOVE,
		DL_CMD_DELETE,
		DL_CMD_PAUSE
	} command;
	enum {
		DL_ST_NONE = 1,
		DL_ST_QUEUE,
		DL_ST_READY,
		DL_ST_RUNNING,
		DL_ST_STOPED,
		DL_ST_PAUSED,
		DL_ST_COMPLETED,
		DL_ST_FREED
	} flag;
	int ref_count;
	int received;
	float speed;
	char *path;				/* absolute path */
	char *status;
	MP3 *item;
} Downitem;

typedef struct upload_item {
	enum {
		UL_READY = 1,
		UL_START,
		UL_ERROR,
		UL_COMPLETE
	} flag;
	char *path;
	MP3 *item;
	MTP *mtp;
} Upitem;


typedef enum {
	NO_SOUND = 0,
	BEEP_SOUND,
	ESD_SOUND
} Sound_Device;

typedef struct application {
	gchar *username;			/* Ҹٴ ̵ */
	gchar *password;			/* Ҹٴ н */
	gchar *soribada_server;
	gchar *proxy_server;
	gint port;				/* listening port */
	Protocol *pl;				/* soribada  */

	gchar *download_dir;			/* ޴ ġ */
	GList *localmp3_dir;			/*  mp3 丮 */

	gchar *mp3_list_file;			/*   ִ mp3  */
	GList *mp3_list;			/*   ִ mp3  */

	gboolean quit_upload;
	gboolean is_login;			/* α  */
	gboolean stop_search;			/* ˻  */
	gboolean online_list_updated;		/* ˻  */
	gboolean search_in_result;		/* ˻ ɼ */
	gboolean redraw_search_clist;

	GIOChannel *listen_channel;		/* listen udp socket */
	GIOChannel *service_channel;		/* listen tcp socket */
	GIOChannel *message_channel;		/* listen message */
	int user_limit;
	int nconnected;

	int bandwidth_limit;			/* in kbytes/s */
	gboolean use_bandwidth_limit;

	/* online list */
	gint nonlines;
	struct node *onlines[MAX_ONLINES];
	struct timeval time[MAX_TIME_TABLE];

	/* ˻ */
	guint16 query_id;
	guint send_query_handler;
	int nreplies;
	int maxreplies;
	MP3 **replies;

	/* downloads table */
	GList *downloads;
	GList *download_queue;
	
	/* message list */
	GList *messages;

	/* main window */
	guint context_id;
	GtkWidget *window;			/* main window */
	GtkWidget *statusbar;			/* status bar */
	GtkWidget *menu_connect_to_server;	/* menu/file/connect to server*/
	GtkWidget *menu_disconnect;		/* menu/file/disconnect */
	GtkWidget *search_progressbar;
	gchar *mp3player;
	pid_t mp3playerpid;

	/* login window */
	GtkWidget *login_window;		/* login window */
	GtkWidget *username_entry;		/* username entry */
	GtkWidget *password_entry;		/* password entry */

	/* search tab */
	GtkWidget *search_entry;
	GtkWidget *search_inverse_entry;
	GtkWidget *search_button;
	GtkWidget *search_clist;		/* ٸ ༮   */

	/* download tab */
	GtkWidget *downloading_clist;		/* downloding list */
	GtkWidget *downloading_clist_resume_menu;
	GtkWidget *downloading_clist_queueing_menu;
	GtkWidget *downloaded_clist;		/* downloded list */

	/* upload tab */
	GtkWidget *upload_clist;
	GtkWidget *query_text;

	/* local tab */
	GtkWidget *local_clist;

	/* message window */
	GtkWidget *message_window;
	GtkWidget *message_clist;
	GtkWidget *message_text;

	/* sound */
	Sound_Device sound_device;

	gchar *sound_on_search;
	gboolean use_sound_on_search;
	gchar *sound_on_down;
	gboolean use_sound_on_down;
	gchar *sound_on_message;
	gboolean use_sound_on_message;

	/* preference */
	guint width;
	guint height;
	gboolean auto_login;
	gboolean save_username;
	gboolean save_password;
	gboolean use_default_server;
	gboolean show_query_list;
	gboolean check_online;
	gboolean use_proxy;

	gboolean search_clist_filename;
	gboolean search_clist_title;
	gboolean search_clist_artist;
	gboolean search_clist_comment;
	gboolean search_clist_size;
	gboolean search_clist_bitrate;
	gboolean search_clist_time;
	gboolean search_clist_user;
	gboolean search_clist_speed;

	gboolean downloading_clist_filename;
	gboolean downloading_clist_user;
	gboolean downloading_clist_size;
	gboolean downloading_clist_received;
	gboolean downloading_clist_progress;
	gboolean downloading_clist_speed;

	gboolean downloaded_clist_filename;
	gboolean downloaded_clist_user;
	gboolean downloaded_clist_size;
	gboolean downloaded_clist_received;
	gboolean downloaded_clist_status;

	gboolean scan_sub_dirs;
} Application;

/* preference */
struct preferences {
	gboolean lock;
	gchar *username;
	gchar *password;
	gchar *soribada_server;
	gchar *proxy_server;
	gchar *download_dir;
	gchar *mp3player;
	GList *localmp3_dir;

	gint port;
	gint user_limit;
	gint maxreplies;
	gint bandwidth_limit;

	guint width;
	guint height;

	gboolean auto_login;
	gboolean save_username;
	gboolean save_password;
	gboolean use_default_server;
	gboolean use_bandwidth_limit;
	gboolean show_query_list;
	gboolean check_online;
	gboolean use_proxy;

	Sound_Device sound_device;
	gchar *sound_on_search;
	gboolean use_sound_on_search;
	gchar *sound_on_down;
	gboolean use_sound_on_down;
	gchar *sound_on_message;
	gboolean use_sound_on_message;

	gboolean search_clist_filename;
	gboolean search_clist_title;
	gboolean search_clist_artist;
	gboolean search_clist_comment;
	gboolean search_clist_size;
	gboolean search_clist_bitrate;
	gboolean search_clist_time;
	gboolean search_clist_user;
	gboolean search_clist_speed;

	gboolean downloading_clist_filename;
	gboolean downloading_clist_user;
	gboolean downloading_clist_size;
	gboolean downloading_clist_received;
	gboolean downloading_clist_progress;
	gboolean downloading_clist_speed;

	gboolean downloaded_clist_filename;
	gboolean downloaded_clist_user;
	gboolean downloaded_clist_size;
	gboolean downloaded_clist_received;
	gboolean downloaded_clist_status;

	gboolean scan_sub_dirs;
};

typedef struct preferences Preferences;

/*   */
extern Application *app;	/* app    struct */
extern Preferences *pref;	/* preferences window    */

/* conf.c */
void load_default_value(void);
void load_conf_file(void);
void write_conf_file(void);

/* misc.c */
int messagebox(const char *fmt, ...);
void *xmalloc(size_t size);
char *basename_for_windows(char *path);
gboolean is_file_exist(const char *filename);
int get_file_size(const char *filename);
gint compare_by_num(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2);
int writen(int fd, const void *vptr, int n);
void remove_crlf_from_str(char *str);
char *xstrstr(const char *haystack, const char *needle);

#endif /* __SORIBADA_H_ */
