/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <glib.h>

#ifdef HAVE_ESD
#include <esd.h>
#endif

#include "soribada.h"
#include "core.h"

static char *which_sound(Sound_Event event) {
	switch (event) {
	case SOUND_ON_SEARCH:
		return app->sound_on_search;
	case SOUND_ON_DOWN:
		return app->sound_on_down;
	case SOUND_ON_MESSAGE:
		return app->sound_on_message;
	default:
		return NULL;
	}
}

void sound_event(Sound_Event event) {
	if (app->sound_device != NO_SOUND) {
		gchar *file = which_sound(event);

		if (file == NULL)
			return;

		switch (app->sound_device) {
		case BEEP_SOUND:
			g_print("\a");
			break;
		case ESD_SOUND:
#ifdef HAVE_ESD
			esd_play_file(NULL, file, 1);
#endif
			break;
		default:
			/* do nothing */
		}
	}
}

void play_wave_file(const gchar *filename, Sound_Device device) {
	switch (device) {
	case BEEP_SOUND:
		g_print("\a");
		break;
	case ESD_SOUND:
#ifdef HAVE_ESD
		esd_play_file(NULL, filename, 1);
#endif
		break;
	default:
		/* do nothing */
	}
}
